import os
import subprocess

from schnetpack.md.utils import MDUnits


class MDCalculatorError(Exception):
    """
    Exception for MDCalculator base class.
    """

    pass


class MDCalculator:
    """
    Base MDCalculator class for computing and updating the forces of the simulated system, as well as other
    properties predicted by the model.

    Args:
        required_properties (list): List of the property names which will be passed to the simulator
        force_handle (str): Name of the property corresponding to the forces.
        position_conversion (float): Unit conversion for the length used in the model computing all properties. E.g. if
                             the model needs Angstrom, one has to provide the conversion factor converting from the
                             atomic units used internally (Bohr) to Angstrom: 0.529177...
        force_conversion (float): Conversion factor converting the forces returned by the used model back to atomic
                                  units (Hartree/Bohr).
        property_conversion (dict(float)): Optional dictionary of conversion factors for other properties predicted by
                                           the model. Only changes the units used for logging the various outputs.
        detach (bool): Detach property computation graph after every calculator call. Enabled by default. Should only
                       be disabled if one wants to e.g. compute derivatives over short trajectory snippets.
    """

    def __init__(
        self,
        required_properties,
        force_handle,
        position_conversion=1.0,
        force_conversion=1.0,
        property_conversion={},
        stress_handle=None,
        stress_conversion=1.0,
        detach=True,
    ):
        self.results = {}
        self.force_handle = force_handle
        self.stress_handle = stress_handle
        self.required_properties = required_properties

        # Perform automatic conversion of units
        # internal -> calculator
        self.position_conversion = MDUnits.internal2unit(position_conversion)

        # calculator -> internal
        self.force_conversion = MDUnits.unit2internal(force_conversion)
        self.stress_conversion = MDUnits.unit2internal(stress_conversion)
        self.property_conversion = {
            p: MDUnits.unit2internal(property_conversion[p])
            for p in property_conversion
        }

        # Default conversion (1.0) for all units not set above
        self._init_default_conversion()

        self.detach = detach

    def check_calculator(self, system):
        # TODO:
        #   1) check cell
        #   2) check cutoff region
        #   3) Reset all known NBLS
        #   4) check PBC
        raise NotImplementedError

    def calculate(self, system):
        """
        Main calculator routine, which needs to be implemented individually.
        This routine should take the current system state, perform the appropriate computations to get the forces
        and use them to update the system forces stored in system.forces

        To this end, results should be stored in the dictionary self.results using the keys contained in
        self.required_properties
        Afterwards, the routine self._update_system(system) can be used to update the system state.

        Args:
            system (schnetpack.md.System): System object containing current state of the simulation.
        """
        raise NotImplementedError

    def _init_default_conversion(self):
        """
        Auxiliary routine to initialize default conversion factors (1.0) if no alternatives are given in
        property_conversion upon initializing the calculator.
        """
        for p in self.required_properties:
            if p not in self.property_conversion:
                self.property_conversion[p] = 1.0

    def _update_system(self, system):
        """
        Routine, which looks in self.results for the properties defined in self.required_properties and uses them to
        update the forces and properties of the provided system. If required, reformatting is carried out here.

        Args:
            system (schnetpack.md.System): System object containing current state of the simulation.
        """

        # Collect all requested properties (including forces)
        for p in self.required_properties:
            if p not in self.results:
                raise MDCalculatorError(
                    "Requested property {:s} not in " "results".format(p)
                )
            else:
                # Detach properties if requested
                if self.detach:
                    self.results[p] = self.results[p].detach()

                dim = self.results[p].shape
                system.properties[p] = (
                    self.results[p].view(
                        system.n_replicas, system.n_molecules, *dim[1:]
                    )
                    * self.property_conversion[p]
                )

        # Set the forces for the system (at this point, already detached)
        self._set_system_forces(system)

        # Set stress of the system if requested:
        if self.stress_handle is not None:
            self._set_system_stress(system)

    def _get_system_molecules(self, system):
        """
        Routine to extract positions, atom_types and atom_masks formatted in a manner suitable for schnetpack models
        from the system class. This is done by collapsing the replica and molecule dimension into one batch dimension.

        Args:
            system (schnetpack.md.System): System object containing current state of the simulation.

        Returns:
            torch.FloatTensor: (n_replicas*n_molecules) x n_atoms x 3 tensor holding nuclear positions
            torch.FloatTensor: (n_replicas*n_molecules) x n_atoms tensor holding nuclear charges
            torch.FloatTensor: (n_replicas*n_molecules) x n_atoms binary tensor indicating padded atom dimensions
        """
        positions = (
            system.positions.view(-1, system.max_n_atoms, 3) * self.position_conversion
        )

        cells = system.cells
        pbc = system.pbc
        if system.cells is not None:
            cells = cells.view(-1, 3, 3) * self.position_conversion
            pbc = pbc.view(-1, 3)

        atom_types = system.atom_types.view(-1, system.max_n_atoms)
        atom_masks = system.atom_masks.view(-1, system.max_n_atoms)

        return positions, atom_types, atom_masks, cells, pbc

    def _set_system_forces(self, system):
        """
        Function to reformat and update the forces of the system from the computed forces stored in self.results.
        The string contained in self.force_handle is used as an indicator. The single batch dimension is recast to the
        original replica x molecule dimensions used by the system.

        Args:
            system (schnetpack.md.System): System object containing current state of the simulation.
        """
        forces = self.results[self.force_handle]
        system.forces = (
            forces.view(system.n_replicas, system.n_molecules, system.max_n_atoms, 3)
            * self.force_conversion
        )

    def _set_system_stress(self, system):
        stress = self.results[self.stress_handle]
        system.stress = (
            stress.view(system.n_replicas, system.n_molecules, 3, 3)
            * self.stress_conversion
        )


class QMCalculatorError(Exception):
    """
    Exception for the QM calculator base class
    """

    pass


class QMCalculator(MDCalculator):
    """
    Basic calculator for interfacing quantum chemistry codes with SchNetPack molecular dynamics.



    Calculator for interfacing the ORCA code package with SchNetPack molecular dynamics.
    Requires ORCA to be installed and an input file template.
    This template is a standard ORCA input file, with everything past the specification of coordinate
    format, charge and multiplicity removed (coordinates and final *).
    If desired, a Queuer can be give, which will attempt to send all jobs to a grid engine queue.

    In general, the calculator will take the current System to generate inputs, perform the calculation
    with ORCA, extract data from the ouput file (useing the OrcaParser class) and update the System.

    Args:
        required_properties (list): List of properties which should be extracted from output.
        force_handle (str): Indicator for molecular forces.
        compdir (str): Directory in which computations are performed.
        qm_executable (str): Path to the ORCA executable.
        position_conversion (str/float, optional): Conversion of positions from atomic units.
        force_conversion (str/float, optional): Conversion of forces to atomic units.
        property_conversion (dict, optional): Convert properties to requested units. If left empty, no conversion
                                              is performed.
        adaptive (bool, optional): Specify, whether the calculator should be used for adaptive sampling.
    """

    is_atomistic = []

    def __init__(
        self,
        required_properties,
        force_handle,
        compdir,
        qm_executable,
        position_conversion="Angstrom",
        force_conversion=1.0,
        property_conversion={},
        adaptive=False,
    ):

        super(QMCalculator, self).__init__(
            required_properties,
            force_handle,
            position_conversion=position_conversion,
            force_conversion=force_conversion,
            property_conversion=property_conversion,
        )

        self.qm_executable = qm_executable

        self.compdir = compdir
        if not os.path.exists(self.compdir):
            os.makedirs(compdir)

        # Set the force handle to be an atomistic property
        self.is_atomistic = force_handle

        self.adaptive = adaptive
        self.step = 0

    def calculate(self, system, samples=None):
        """
        Perform the calculation with a quantum chemistry code.
        If samples is given, only a subset of molecules is selected.

        Args:
            system (schnetpack.md.System): System from the molecular dynamics simulation.
            samples (np.array, optional): Integer array specifying whether only particular
                                          replicas and molecules in the system should be used for
                                          computations. Only works with adaptive sampling.

        Returns:
            (list,list):
                atom_buffer:
                    List of ASE atoms objects of every computed molecule.
                    Only returned if adaptive sampling is activated.

                property_buffer:
                    List of property dictionaries for every computation.
                    Only returned if adaptive sampling is activated.
        """
        # Use of samples only makes sense in conjunction with adaptive sampling
        if not self.adaptive and samples is not None:
            raise QMCalculatorError(
                "Usage of subsamples only allowed during adaptive sampling."
            )

        # Generate director for current step
        # current_compdir = os.path.join(self.compdir, 'step_{:06d}'.format(self.step))
        current_compdir = os.path.join(self.compdir, "step_X")
        if not os.path.exists(current_compdir):
            os.makedirs(current_compdir)

        # Get molecules (select samples if requested)
        molecules = self._extract_molecules(system, samples=samples)

        # Run computation
        outputs = self._run_computation(molecules, current_compdir)

        # Increment internal step
        self.step += 1

        # Prepare output
        # a) either parse to update system properties
        if not self.adaptive:
            self.results = self._format_calc(outputs, system)
            self._update_system(system)
        # b) or append to the database (just return everything as molecules/atoms objects)
        else:
            atom_buffer, property_buffer = self._format_ase(molecules, outputs)
            return atom_buffer, property_buffer

    def _extract_molecules(self, system, samples=None):
        """
        Extract atom types and molecular structures from the system. and convert to
        appropriate units.

        Args:
            system (schnetpack.md.System): System from the molecular dynamics simulation.
            samples (np.array, optional): Integer array specifying whether only particular
                                          replicas and molecules in the system should be used for
                                          computations. Only works with adaptive sampling.

        Returns:
            list: List of tuples containing the atom types (integer numpy.array) and positions
                  (float numpy.array).
        """
        molecules = []
        for rep_idx in range(system.n_replicas):
            for mol_idx in range(system.n_molecules):
                # Check which geometries need samples in adaptive setup
                if samples is not None:
                    if not samples[rep_idx, mol_idx]:
                        continue
                atom_types = system.atom_types[
                    rep_idx, mol_idx, : system.n_atoms[mol_idx]
                ]
                # Convert Bohr to Angstrom
                positions = (
                    system.positions[rep_idx, mol_idx, : system.n_atoms[mol_idx], ...]
                    * self.position_conversion
                )
                # Store atom types and positions for ase db during sampling
                molecules.append((atom_types, positions))

        return molecules

    def _run_computation(self, molecules, current_compdir):
        """
        Placeholder performing the computation.

        Args:
            molecules (list): List of tuples containing the atom types (integer numpy.array)
                      and positions (float numpy.array).
            current_compdir (str): Path to the current computation directory.
        """
        raise NotImplementedError

    def _format_calc(self, outputs, system):
        """
        Placeholder to format the computation output if no adaptive sampling is used.

        Args:
            outputs (list): Paths to output files.
            system (schnetpack.md.System): System from the molecular dynamics simulation.
        """
        raise NotImplementedError

    def _format_ase(self, molecules, outputs):
        """
        Placeholder to format the ouput for storage in an ASE database (for adaptive sampling).

        Args:
            molecules (list): List of tuples containing the atom types (integer numpy.array)
                      and positions (float numpy.array).
            outputs (list): Paths to output files.
        """
        raise NotImplementedError


class Queuer:
    """
    Base class for interfacing a calculator with a grid engine queue.
    Will create the directory structure and a submission command, which it will pass to the
    given queue. Jobs will be submitted as array jobs, where the number of concurrently running
    jobs can be changed. In order to the script to work, the QUEUE_FILE string needs to be updated
    and the _create_submission_command function has to be implemented.

    Args:
        queue (str): Indentifier for the queue the job should be sent to.
        executable (str): Path to the executable to be called.
        concurrent (str): How many concurrent jobs should be run in each array. (default=100)
        basename (str): Basic identifier used for job (default='input')
        cleanup (bool): Whether directories should be deleted (default=True)
    """

    QUEUE_FILE = """
#!/usr/bin/env bash
##############################
#$ -cwd
#$ -V
#$ -q {queue}
#$ -N {jobname}
#$ -t 1-{array_range}
#$ -tc {concurrent}
#$ -S /bin/bash
#$ -e /dev/null
#$ -o /dev/null
#$ -r n
#$ -sync y
##############################

# Adapt here
"""

    def __init__(
        self, queue, executable, concurrent=100, basename="input", cleanup=True
    ):
        self.queue = queue
        self.executable = executable
        self.concurrent = concurrent
        self.basename = basename
        self.cleanup = cleanup

    def submit(self, input_files, current_compdir):
        """
        Submit an array of jobs to a grid engine and wait for their completion.

        Args:
            input_files (list): List of all inputs.
            current_compdir (str): Path to the directory used for computations.
        """
        jobname = os.path.basename(current_compdir)
        compdir = os.path.abspath(current_compdir)
        n_inputs = len(input_files)

        submission_command = self._create_submission_command(n_inputs, compdir, jobname)

        script_name = os.path.join(current_compdir, "submit.sh")
        with open(script_name, "w") as submission_script:
            submission_script.write(submission_command)

        computation = subprocess.Popen(["qsub", script_name], stdout=subprocess.PIPE)
        computation.wait()

        if self.cleanup:
            os.remove(script_name)

    def _create_submission_command(self, n_inputs, compdir, jobname):
        """
        Creates the submission command and needs to be implemeneted for specific tasks.

        Args:
            n_inputs (int): Number of input files to be submitted
            compdir (str): Current working directory
            jobname (str): Name of each job

        Returns:
            str: Submission command.
        """
        raise NotImplementedError



"""
Classes for constructing the different representations available in SchnetPack. This encompasses SchNet [#schnet4]_,
Behler-type atom centered symmetry functions (ACSF) [#acsf2]_ and a weighted variant thereof (wACSF) [#wacsf2]_.

References
----------
.. [#schnet4] Schütt, Arbabzadah, Chmiela, Müller, Tkatchenko:
   Quantum-chemical insights from deep tensor neural networks.
   Nature Communications, 8, 13890. 2017.
.. [#acsf2] Behler:
   Atom-centered symmetry functions for constructing high-dimensional neural network potentials.
   The Journal of Chemical Physics 134. 074106. 2011.
.. [#wacsf2] Gastegger, Schwiedrzik, Bittermann, Berzsenyi, Marquetand:
   wACSF -- Weighted atom-centered symmetry functions as descriptors in machine learning potentials.
   The Journal of Chemical Physics 148 (24), 241709. 2018.
"""

from schnetpack.representation.schnet import *
from schnetpack.representation.hdnn import *



"""
All logging operations in SchNetPack molecular dynamics simulations are performed via
simulation hooks. This includes the generation of checkpoint files. The main tool to
store simulation data is the :obj:`schnetpack.md.simulation_hooks.FileLogger`, which uses
data streams to collect information (positions, velocities, properties, ...) during a
simulation and store it to specially formatted HDF5 files. These files can then be read
using the :obj:`schnetpack.md.utils.HDF5Loader`.
"""

import json
import os
import h5py
import numpy as np
import torch

from schnetpack.md.simulation_hooks import SimulationHook
import schnetpack as spk

__all__ = [
    "Checkpoint",
    "TemperatureLogger",
    "PressureLogger",
    "FileLogger",
    "MoleculeStream",
    "DataStream",
    "PropertyStream",
    "SimulationStream",
]


class Checkpoint(SimulationHook):
    """
    Hook for writing out checkpoint files containing the state_dict of the simulator. Used to restart the simulation
    from a previous step of previous system configuration.

    Args:
        checkpoint_file (str): Name of the file used to store the state_dict periodically.
        every_n_steps (int): Frequency with which checkpoint files are written.
    """

    def __init__(self, checkpoint_file, every_n_steps=1000):
        super(Checkpoint, self).__init__()
        self.every_n_steps = every_n_steps
        self.checkpoint_file = checkpoint_file

    def on_step_end(self, simulator):
        """
        Store state_dict at specified intervals.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        if simulator.step % self.every_n_steps == 0:
            torch.save(simulator.state_dict, self.checkpoint_file)

    def on_simulation_end(self, simulator):
        """
        Store state_dict at the end of the simulation.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        torch.save(simulator.state_dict, self.checkpoint_file)


class DataStream:
    """
    Basic DataStream class to be used with the FileLogger. Creates data groups in the main hdf5 file, accumulates
    the associated information and flushes them to the file periodically.

    Args:
        group_name (str): Name of the data group entry.
    """

    def __init__(self, group_name):
        self.group_name = group_name
        self.buffer = None
        self.data_group = None

        self.main_dataset = None
        self.buffer_size = None
        self.restart = None
        self.every_n_steps = None

    def init_data_stream(
        self, simulator, main_dataset, buffer_size, restart=False, every_n_steps=1
    ):
        """
        Wrapper for initializing the data containers based on the instructions provided in the current simulator. For
        every data stream, the current number of valid entries is stored, which is updated periodically. This is
        necessary if a simulation is e.g. restarted or data is extracted during a running simulations, as all arrays
        are initially constructed taking the full length of the simulation into account.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
            main_dataset (h5py.File): Main h5py dataset object.
            buffer_size (int): Size of the buffer, once full, data is stored to the hdf5 dataset.
            restart (bool): If the simulation is restarted, continue logging in the previously created dataset.
                            (default=False)
            every_n_steps (int): How often simulation steps are logged. Used e.g. to determine overall time step in
                                 MoleculeStream.
        """
        self.main_dataset = main_dataset
        self.buffer_size = buffer_size
        self.restart = restart
        self.every_n_steps = every_n_steps

        self._init_data_stream(simulator)
        # Write number of meaningful entries into attributes
        if not self.restart:
            self.data_group.attrs["entries"] = 0

    def _init_data_stream(self, simulator):
        """
        Specific initialization routine. Needs to be adapted.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        raise NotImplementedError

    def update_buffer(self, buffer_position, simulator):
        """
        Instructions for updating the buffer. Needs to take into account reformatting of data, etc.

        Args:
            buffer_position (int): Current position in the buffer.
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        raise NotImplementedError

    def flush_buffer(self, file_position, buffer_position):
        """
        Write data contained in buffer into the main hdf5 file.

        Args:
            file_position (int): Current position in the main dataset file.
            buffer_position (int): Most recent entry in the buffer. Used to ensure no buffer entries are written to the
                                   main file.
        """
        self.data_group[file_position : file_position + buffer_position] = (
            self.buffer[:buffer_position].detach().cpu()
        )
        # Update number of meaningful entries
        self.data_group.attrs.modify("entries", file_position + buffer_position)
        self.data_group.flush()

    def _setup_data_groups(self, data_shape, simulator):
        """
        Auxiliary routine for initializing data groups in the main hdf5 data file as well as the buffer used during
        logging. All arrays are initialized using the full number of simulation steps specified in the main simulator
        class. The current positions in these arrays are managed via the 'entries' group attribute.

        Args:
            data_shape (list(int)): Shape of the target data tensor
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        # Initialize the buffer
        self.buffer = torch.zeros(
            self.buffer_size, *data_shape, device=simulator.system.device
        )

        if self.restart:
            # Load previous data stream and resize
            self.data_group = self.main_dataset[self.group_name]
            self.data_group.resize(
                (simulator.n_steps + self.data_group.attrs["entries"],) + data_shape
            )
        else:
            # Otherwise, generate new data group in the dataset
            self.data_group = self.main_dataset.create_dataset(
                self.group_name,
                shape=(simulator.n_steps,) + data_shape,
                chunks=self.buffer.shape,
                dtype=np.float32,
                maxshape=(None,) + data_shape,
            )


class PropertyStream(DataStream):
    """
    Main routine for logging the properties predicted by the calculator to the group 'properties' of hdf5 dataset.
    Stores properties in a flattened array and writes names, shapes and positions to the group data section. Since this
    routine determines property shapes based on the system.properties dictionary, at least one computations needs to be
    performed beforehand. Properties are stored in an array of the shape
    n_steps x n_replicas x n_molecules x n_properties, where n_steps is the number of simulation steps, n_replicas and
    n_molecules is the number of simulation replicas and different molecules and n_properties is the length of the
    flattened property array.

    Args:
        target_properties (list): List of properties to be written to the hdf5 database. If no list is given, defaults
                                  to None, which means all properties are stored.
    """

    def __init__(self, target_properties=None):
        super(PropertyStream, self).__init__("properties")
        self.n_replicas = None
        self.n_molecules = None
        self.properties_slices = {}
        self.target_properties = target_properties

    def _init_data_stream(self, simulator):
        """
        Routine for determining the present properties and their respective shapes based on the
        simulator.system.properties dictionary and storing them into the attributes of the hdf5 data group.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        self.n_replicas = simulator.system.n_replicas
        self.n_molecules = simulator.system.n_molecules

        if simulator.system.properties is None:
            raise FileLoggerError(
                "Shape of properties could not be determined," "please call calculator"
            )

        # Determine present properties, order and shape thereof
        (
            properties_entries,
            properties_shape,
            properties_positions,
        ) = self._get_properties_structures(simulator.system.properties)

        data_shape = (self.n_replicas, self.n_molecules, properties_entries)

        self._setup_data_groups(data_shape, simulator)

        if not self.restart:
            # Store metadata on shape and position of properties in array
            self.data_group.attrs["shapes"] = json.dumps(properties_shape)
            self.data_group.attrs["positions"] = json.dumps(properties_positions)

    def update_buffer(self, buffer_position, simulator):
        """
        Routine for updating the propery buffer.

        Args:
            buffer_position (int): Current position in the buffer.
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        # These are already detached in the calculator by default.
        for p in self.properties_slices:
            self.buffer[
                buffer_position : buffer_position + 1, ..., self.properties_slices[p]
            ] = (
                simulator.system.properties[p]
                .contiguous()
                .view(self.n_replicas, self.n_molecules, -1)
            )
            # self.buffer[buffer_position:buffer_position + 1, ..., self.properties_slices[p]] = \
            #     simulator.system.properties[p].view(self.n_replicas, self.n_molecules, -1).detach()

    def _get_properties_structures(self, property_dict):
        """
        Auxiliary function to get the names, shapes and positions used in the property stream based on the property
        dictionary of the system.

        Args:
            property_dict (dict(torch.Tensor)): Property dictionary of the main simulator.system class.

        Returns:
            int: Total number of property fields used per replica, molecule and time step.
            dict(slice): Dictionary holding the position of the target property within the flattened array.
            dist(tuple): Dictionary holding the original shapes of the property tensors.
        """
        properties_entries = 0
        properties_shape = {}
        properties_positions = {}

        for p in property_dict:

            if self.target_properties is not None and p not in self.target_properties:
                continue

            # Store shape for metadata
            properties_shape[p] = [int(i) for i in property_dict[p].shape[2:]]
            # Use shape to determine overall array dimensions
            start = properties_entries
            properties_entries += int(np.prod(properties_shape[p]))
            # Get position of property in array
            properties_positions[p] = (start, properties_entries)
            self.properties_slices[p] = slice(start, properties_entries)

        return properties_entries, properties_shape, properties_positions


class SimulationStream(PropertyStream):
    """
    DataStream for dynamic system properties, such as kinetic energy and temperatures of the individual replicas, as
    well as centroids. Adds and updates the group 'simulation'. Stores properties in a flattened array and writes names,
    shapes and positions to the group data section. Properties are stored in an array of the shape
    n_steps x n_replicas x n_molecules x n_properties, where n_steps is the number of simulation steps, n_replicas and
    n_molecules is the number of simulation replicas and different molecules and n_properties is the length of the
    flattened property array.
    """

    def __init__(self):
        super(SimulationStream, self).__init__()
        self.group_name = "simulation"

    def _init_data_stream(self, simulator):
        """
        Get the shape and positions of all monitored properties.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        self.n_replicas = simulator.system.n_replicas
        self.n_molecules = simulator.system.n_molecules

        # Create an auxiliary property array from the current system state
        property_dictionary = {
            "kinetic_energy": simulator.system.kinetic_energy,
            "kinetic_energy_centroid": simulator.system.centroid_kinetic_energy,
            "temperature": simulator.system.temperature,
            "temperature_centroid": simulator.system.centroid_temperature,
        }

        (
            properties_entries,
            properties_shape,
            properties_positions,
        ) = self._get_properties_structures(property_dictionary)

        data_shape = (self.n_replicas, self.n_molecules, properties_entries)

        self._setup_data_groups(data_shape, simulator)

        if not self.restart:
            self.data_group.attrs["shapes"] = json.dumps(properties_shape)
            self.data_group.attrs["positions"] = json.dumps(properties_positions)

    def update_buffer(self, buffer_position, simulator):
        """
        Routine for updating the buffer.

        Args:
            buffer_position (int): Current position in the buffer.
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        property_dictionary = {
            "temperature": simulator.system.temperature,
            "kinetic_energy": simulator.system.kinetic_energy,
            "temperature_centroid": simulator.system.centroid_temperature,
            "kinetic_energy_centroid": simulator.system.centroid_kinetic_energy,
        }

        for p in self.properties_slices:
            self.buffer[
                buffer_position : buffer_position + 1, ..., self.properties_slices[p]
            ] = property_dictionary[p].view(-1, self.n_molecules, 1)


class MoleculeStream(DataStream):
    """
    DataStream for logging atom types, positions and velocities to the group 'molecules' of the main hdf5 dataset.
    Positions and velocities are stored in a n_steps x n_replicas x n_molecules x 6 array, where n_steps is the number
    of simulation steps, n_replicas and n_molecules are the number of simulation replicas and different molecules. The
    first 3 of the final 6 components are the Cartesian positions and the last 3 the velocities in atomic units. Atom
    types, the numbers of replicas, molecules and atoms, as well as the length of the time step in atomic units
    (for spectra) are stored in the group attributes.
    """

    def __init__(self):
        super(MoleculeStream, self).__init__("molecules")
        self.written = 0

    def _init_data_stream(self, simulator):
        """
        Initialize the main data shape and write information on atom types, the numbers of replicas, molecules and
        atoms, as well as the length of the time step in atomic units to the group attributes.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        # Get shape of array depending on whether forces should be stored.

        if simulator.system.cells is not None:
            data_shape = (
                simulator.system.n_replicas,
                simulator.system.n_molecules,
                simulator.system.max_n_atoms * 6 + 9,
            )
        else:
            data_shape = (
                simulator.system.n_replicas,
                simulator.system.n_molecules,
                simulator.system.max_n_atoms * 6,
            )

        self._setup_data_groups(data_shape, simulator)

        if not self.restart:
            self.data_group.attrs["n_replicas"] = simulator.system.n_replicas
            self.data_group.attrs["n_molecules"] = simulator.system.n_molecules
            self.data_group.attrs["max_n_atoms"] = simulator.system.max_n_atoms
            self.data_group.attrs["n_atoms"] = simulator.system.n_atoms.cpu()
            self.data_group.attrs["atom_types"] = simulator.system.atom_types.cpu()
            self.data_group.attrs["pbc"] = simulator.system.pbc.cpu()

            self.data_group.attrs["time_step"] = (
                simulator.integrator.time_step * self.every_n_steps
            )
            self.data_group.attrs["every_n_steps"] = self.every_n_steps

    def update_buffer(self, buffer_position, simulator):
        """
        Routine for updating the buffer.

        Args:
            buffer_position (int): Current position in the buffer.
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        n_entries = simulator.system.max_n_atoms * 3

        self.buffer[
            buffer_position : buffer_position + 1, ..., :n_entries
        ] = simulator.system.positions.view(
            simulator.system.n_replicas, simulator.system.n_molecules, -1
        )
        self.buffer[
            buffer_position : buffer_position + 1, ..., n_entries : 2 * n_entries
        ] = simulator.system.velocities.view(
            simulator.system.n_replicas, simulator.system.n_molecules, -1
        )

        if simulator.system.cells is not None:
            self.buffer[
                buffer_position : buffer_position + 1, ..., 2 * n_entries :
            ] = simulator.system.cells.view(
                simulator.system.n_replicas, simulator.system.n_molecules, -1
            )


class FileLoggerError(Exception):
    """
    Exception for the FileLogger class.
    """

    pass


class FileLogger(SimulationHook):
    """
    Class for monitoring the simulation and storing the resulting data to a hfd5 dataset. The properties to monitor are
    given via instances of the DataStream class. Uses buffers of a given size, which are accumulated and fushed to the
    main file in regular intervals in order to reduce I/O overhead. All arrays are initialized for the full number of
    requested simulation steps, the current positions in each data group is handled via the 'entries' attribute.

    Args:
        filename (str): Path to the hdf5 database file.
        buffer_size (int): Size of the buffer, once full, data is stored to the hdf5 dataset.
        data_streams list(schnetpack.simulation_hooks.hooks.DataStream): List of DataStreams used to collect and log information
                                                                 to the main hdf5 dataset, default are properties and
                                                                 molecules.
        every_n_steps (int): Frequency with which the buffer is updated.
    """

    def __init__(
        self,
        filename,
        buffer_size,
        data_streams=[MoleculeStream(), PropertyStream()],
        every_n_steps=1,
    ):

        self.every_n_steps = every_n_steps
        self.filename = filename
        self.buffer_size = buffer_size

        # Create an empty variable to hold the HDF5 file upon initialization
        self.file = None

        # Precondition data streams
        self.data_steams = []
        for stream in data_streams:
            self.data_steams += [stream]

        # Counter for file writes
        self.file_position = 0
        self.buffer_position = 0

    def on_simulation_start(self, simulator):
        """
        Initializes all present data streams (creating groups, determining buffer shapes, storing metadata, etc.). In
        addition, the 'entries' attribute of each data stream is read from the existing data set upon restart.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """

        # Flag, if new database should be started or data appended to old one
        append_data = False

        # Check, whether file already exists
        if os.path.exists(self.filename):

            # If file exists and it is the first call of a simulator without restart,
            # raise and error.
            if (not simulator.restart) and (simulator.effective_steps == 0):
                raise FileLoggerError(
                    "File {:s} already exists and simulation was not restarted.".format(
                        self.filename
                    )
                )

            # If either a restart is requested or the simulator has already been called,
            # append to file if it exists.
            if simulator.restart or (simulator.effective_steps > 0):
                append_data = True
        else:
            # If no file is found, automatically generate new one.
            append_data = False

        # Create the HDF5 file
        self.file = h5py.File(self.filename, "a", libver="latest")

        # Construct stream buffers and data groups
        for stream in self.data_steams:
            stream.init_data_stream(
                simulator,
                self.file,
                self.buffer_size,
                restart=append_data,
                every_n_steps=self.every_n_steps,
            )

            # Upon restart, get current position in file
            if append_data:
                self.file_position = stream.data_group.attrs["entries"]

        # Enable single writer, multiple reader flag
        self.file.swmr_mode = True

    def on_step_end(self, simulator):
        """
        Update the buffer of each stream after each specified interval and flush the buffer to the main file if full.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        if simulator.step % self.every_n_steps == 0:
            # If buffers are full, write to file
            if self.buffer_position == self.buffer_size:
                self._write_buffer()

            # Update stream buffers
            for stream in self.data_steams:
                stream.update_buffer(self.buffer_position, simulator)

            self.buffer_position += 1

    def on_simulation_end(self, simulator):
        """
        Perform one final flush of the buffers and close the file upon the end of the simulation.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        # Flush remaining data in buffer
        if self.buffer_position != 0:
            self._write_buffer()

        # Close database file
        self.file.close()

    def _write_buffer(self):
        """
        Write all current buffers to the database file.
        """
        for stream in self.data_steams:
            stream.flush_buffer(self.file_position, self.buffer_position)

        self.file_position += self.buffer_size
        self.buffer_position = 0


class TensorboardLogger(SimulationHook):
    """
    Class for logging scalar information of the system replicas and molecules collected during the simulation to
    TensorBoard. An individual scalar is created for every molecule, replica and property.

    Args:
        log_file (str): Path to the TensorBoard file.
        every_n_steps (int): Frequency with which data is logged to TensorBoard.
    """

    def __init__(self, log_file, every_n_steps=100):
        from tensorboardX import SummaryWriter

        self.log_file = log_file
        self.every_n_steps = every_n_steps
        self.writer = SummaryWriter(self.log_file)

        self.n_replicas = None
        self.n_molecules = None

    def on_simulation_start(self, simulator):
        """
        Extract the number of molecules and replicas from simulator.system upon simulation start.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        self.n_replicas = simulator.system.n_replicas
        self.n_molecules = simulator.system.n_molecules

    def on_step_end(self, simulator):
        """
        Routine for collecting and storing scalar properties of replicas and molecules during the simulation. Needs to
        be adapted based on the properties.
        In the easiest case, information on group names, etc. is passed to the self._log_group auxiliary function.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        raise NotImplementedError

    def _log_group(self, group_name, step, property, property_centroid=None):
        """
        Auxiliary routine for logging the scalar data associated with the target property. An individual entry is
        created for every replica and molecule. If requested, an entry corresponding to the systems centroid is also
        created.

        Args:
            group_name (str): Base name of the property group to log.
            step (int): Current simulation step.
            property (torch.Tensor): Tensor of the shape (n_replicas x n_molecules) holding the scalar properties of
                                     each replica and molecule.
            property_centroid (torch.Tensor): Also store the centroid of the monitored property if provided
                                              (default=None).
        """
        logger_dict = {}

        for molecule in range(self.n_molecules):
            mol_name = "{:s}/molecule_{:02d}".format(group_name, molecule + 1)

            if property_centroid is not None:
                logger_dict["centroid"] = property_centroid[0, molecule]

            for replica in range(self.n_replicas):
                rep_name = "r{:02d}".format(replica + 1)
                logger_dict[rep_name] = property[replica, molecule]

            self.writer.add_scalars(mol_name, logger_dict, step)

    def on_simulation_end(self, simulator):
        """
        Close the TensorBoard logger.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        self.writer.close()


class TemperatureLogger(TensorboardLogger):
    """
    TensorBoard logging hook for the temperatures of the replicas, as well as of the corresponding centroids for each
    molecule in the system container.

    Args:
        log_file (str): Path to the TensorBoard file.
        every_n_steps (int): Frequency with which data is logged to TensorBoard.
    """

    def __init__(self, log_file, every_n_steps=100):
        super(TemperatureLogger, self).__init__(log_file, every_n_steps=every_n_steps)
        self.initial_value = None

    def on_step_end(self, simulator):
        """
        Log the systems temperatures at the given intervals.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """
        if simulator.step % self.every_n_steps == 0:
            # Use the _log_group routine to log the systems temperatures
            self._log_group(
                "temperature",
                simulator.step,
                simulator.system.temperature,
                property_centroid=simulator.system.centroid_temperature,
            )


class PressureLogger(TensorboardLogger):
    """
    TensorBoard logging hook for the pressure of the replicas, as well as of the corresponding centroids for each
    molecule in the system container. This requires the stress tensor to be computed in the model.

    Args:
        log_file (str): Path to the TensorBoard file.
        every_n_steps (int): Frequency with which data is logged to TensorBoard.
    """

    def __init__(self, log_file, every_n_steps=100):
        super(PressureLogger, self).__init__(log_file, every_n_steps=every_n_steps)
        self.initial_value = None

    def on_step_end(self, simulator):
        """
        Log the systems pressures at the given intervals.

        Args:
            simulator (schnetpack.simulation_hooks.Simulator): Simulator class used in the molecular dynamics simulation.
        """

        if simulator.step % self.every_n_steps == 0:

            # Log the pressure in bar
            pressure = (
                simulator.system.compute_pressure(kinetic_component=True)
                / spk.md.MDUnits.bar2internal
            )
            centroid_pressure = torch.mean(pressure, 0, keepdim=True)

            # Use the _log_group routine to log the systems temperatures
            self._log_group(
                "pressure",
                simulator.step,
                pressure,
                property_centroid=centroid_pressure,
            )



import torch
from torch import nn

from schnetpack import Properties
from schnetpack.nn import shifted_softplus, Dense

__all__ = ["MLP", "TiledMultiLayerNN", "ElementalGate", "GatedNetwork"]


class MLP(nn.Module):
    """Multiple layer fully connected perceptron neural network.

    Args:
        n_in (int): number of input nodes.
        n_out (int): number of output nodes.
        n_hidden (list of int or int, optional): number hidden layer nodes.
            If an integer, same number of node is used for all hidden layers resulting
            in a rectangular network.
            If None, the number of neurons is divided by two after each layer starting
            n_in resulting in a pyramidal network.
        n_layers (int, optional): number of layers.
        activation (callable, optional): activation function. All hidden layers would
            the same activation function except the output layer that does not apply
            any activation function.

    """

    def __init__(
        self, n_in, n_out, n_hidden=None, n_layers=2, activation=shifted_softplus
    ):
        super(MLP, self).__init__()
        # get list of number of nodes in input, hidden & output layers
        if n_hidden is None:
            c_neurons = n_in
            self.n_neurons = []
            for i in range(n_layers):
                self.n_neurons.append(c_neurons)
                c_neurons = max(n_out, c_neurons // 2)
            self.n_neurons.append(n_out)
        else:
            # get list of number of nodes hidden layers
            if type(n_hidden) is int:
                n_hidden = [n_hidden] * (n_layers - 1)
            self.n_neurons = [n_in] + n_hidden + [n_out]

        # assign a Dense layer (with activation function) to each hidden layer
        layers = [
            Dense(self.n_neurons[i], self.n_neurons[i + 1], activation=activation)
            for i in range(n_layers - 1)
        ]
        # assign a Dense layer (without activation function) to the output layer
        layers.append(Dense(self.n_neurons[-2], self.n_neurons[-1], activation=None))
        # put all layers together to make the network
        self.out_net = nn.Sequential(*layers)

    def forward(self, inputs):
        """Compute neural network output.

        Args:
            inputs (torch.Tensor): network input.

        Returns:
            torch.Tensor: network output.

        """
        return self.out_net(inputs)


class TiledMultiLayerNN(nn.Module):
    """
    Tiled multilayer networks which are applied to the input and produce n_tiled different outputs.
    These outputs are then stacked and returned. Used e.g. to construct element-dependent prediction
    networks of the Behler-Parrinello type.

    Args:
        n_in (int): number of input nodes
        n_out (int): number of output nodes
        n_tiles (int): number of networks to be tiled
        n_hidden (int): number of nodes in hidden nn (default 50)
        n_layers (int): number of layers (default: 3)
    """

    def __init__(
        self, n_in, n_out, n_tiles, n_hidden=50, n_layers=3, activation=shifted_softplus
    ):
        super(TiledMultiLayerNN, self).__init__()
        self.mlps = nn.ModuleList(
            [
                MLP(
                    n_in,
                    n_out,
                    n_hidden=n_hidden,
                    n_layers=n_layers,
                    activation=activation,
                )
                for _ in range(n_tiles)
            ]
        )

    def forward(self, inputs):
        """
        Args:
            inputs (torch.Tensor): Network inputs.

        Returns:
            torch.Tensor: Tiled network outputs.

        """
        return torch.cat([net(inputs) for net in self.mlps], 2)


class ElementalGate(nn.Module):
    """
    Produces a Nbatch x Natoms x Nelem mask depending on the nuclear charges passed as an argument.
    If onehot is set, mask is one-hot mask, else a random embedding is used.
    If the trainable flag is set to true, the gate values can be adapted during training.

    Args:
        elements (set of int): Set of atomic number present in the data
        onehot (bool): Use one hit encoding for elemental gate. If set to False, random embedding is used instead.
        trainable (bool): If set to true, gate can be learned during training (default False)
    """

    def __init__(self, elements, onehot=True, trainable=False):
        super(ElementalGate, self).__init__()
        self.trainable = trainable

        # Get the number of elements, as well as the highest nuclear charge to use in the embedding vector
        self.nelems = len(elements)
        maxelem = int(max(elements) + 1)

        self.gate = nn.Embedding(maxelem, self.nelems)

        # if requested, initialize as one hot gate for all elements
        if onehot:
            weights = torch.zeros(maxelem, self.nelems)
            for idx, Z in enumerate(elements):
                weights[Z, idx] = 1.0
            self.gate.weight.data = weights

        # Set trainable flag
        if not trainable:
            self.gate.weight.requires_grad = False

    def forward(self, atomic_numbers):
        """
        Args:
            atomic_numbers (torch.Tensor): Tensor containing atomic numbers of each atom.

        Returns:
            torch.Tensor: One-hot vector which is one at the position of the element and zero otherwise.

        """
        return self.gate(atomic_numbers)


class GatedNetwork(nn.Module):
    """
    Combines the TiledMultiLayerNN with the elemental gate to obtain element specific atomistic networks as in typical
    Behler--Parrinello networks [#behler1]_.

    Args:
        nin (int): number of input nodes
        nout (int): number of output nodes
        nnodes (int): number of nodes in hidden nn (default 50)
        nlayers (int): number of layers (default 3)
        elements (set of ints): Set of atomic number present in the data
        onehot (bool): Use one hit encoding for elemental gate. If set to False, random embedding is used instead.
        trainable (bool): If set to true, gate can be learned during training (default False)
        activation (callable): activation function

    References
    ----------
    .. [#behler1] Behler, Parrinello:
       Generalized Neural-Network Representation of High-Dimensional Potential-Energy Surfaces.
       Phys. Rev. Lett. 98, 146401. 2007.

    """

    def __init__(
        self,
        nin,
        nout,
        elements,
        n_hidden=50,
        n_layers=3,
        trainable=False,
        onehot=True,
        activation=shifted_softplus,
    ):
        super(GatedNetwork, self).__init__()
        self.nelem = len(elements)
        self.gate = ElementalGate(elements, trainable=trainable, onehot=onehot)
        self.network = TiledMultiLayerNN(
            nin,
            nout,
            self.nelem,
            n_hidden=n_hidden,
            n_layers=n_layers,
            activation=activation,
        )

    def forward(self, inputs):
        """
        Args:
            inputs (dict of torch.Tensor): SchNetPack format dictionary of input tensors.

        Returns:
            torch.Tensor: Output of the gated network.
        """
        # At this point, inputs should be the general schnetpack container
        atomic_numbers = inputs[Properties.Z]
        representation = inputs["representation"]
        gated_network = self.gate(atomic_numbers) * self.network(representation)
        return torch.sum(gated_network, -1, keepdim=True)



"""
Module for extracting information from ORCA output files. Can be used to read arbitrary
information from the output files, as well as gradient and hessian files generated by
the ORCA code package. Containes several predefined parsers.
"""
import logging
import os
import numpy as np
from ase import Atoms, units
from tqdm import tqdm

import schnetpack as spk
from schnetpack import Properties

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))

__all__ = [
    "ppm2au",
    "OrcaParserException",
    "OrcaParser",
    "format_dipole_derivatives",
    "format_polarizability_derivatives",
    "OrcaOutputParser",
    "OrcaFormatter",
    "OrcaPropertyParser",
    "OrcaMainFileParser",
    "OrcaHessianFileParser",
]

# Conversion from ppm to atomic units. Alpha is the fine structure constant and 1e6 are
# the ppm
ppm2au = 1.0 / (units.alpha ** 2 * 1e6)


class OrcaParserException(Exception):
    """
    Exception for OrcaParser class.
    """

    pass


class OrcaParser:
    """
    Main parsing utility for ORCA output files. Runs over a list of output files, extracts the data
    and stores it into a formatted ASE database in SchNetPack format (:obj:`schnetpack.data.AtomsData`).
    This class makes use of the :obj:`OrcaMainFileParser` and :obj:`OrcaHessianFileParser` defined below.

    Args:
        dbpath (str): Path to the target database.
        properties (list): List of properties to extract from ORCA files.
        filter (dict, optional): Dictionary giving the name of a property and a threshold value. Entries in
                                 in the output files with values exceeding the threshold in magnitude are
                                 discarded. This can be used to e.g. screen for numerical noise in implicit
                                 solvent computations, etc.
        mask_charges (bool, optional): If the ORCA calculation used external charges, these are removed from the
                                       positions and atom types read by the parser.
    """

    main_properties = [
        Properties.energy,
        Properties.forces,
        Properties.dipole_moment,
        Properties.polarizability,
        Properties.shielding,
    ]
    hessian_properties = [
        Properties.hessian,
        Properties.dipole_derivatives,
        Properties.polarizability_derivatives,
    ]
    file_extensions = {Properties.forces: ".engrad", Properties.hessian: ".oinp.hess"}
    atomistic = ["atoms", Properties.forces, Properties.shielding]

    def __init__(self, dbpath, properties, filter=None, mask_charges=False):
        self.dbpath = dbpath

        main_properties = []
        hessian_properties = []
        dummy_properties = []

        for p in properties:
            if p in self.main_properties:
                main_properties.append(p)
            elif p in self.hessian_properties:
                hessian_properties.append(p)
            else:
                print("Unrecognized property {:s}".format(p))

        all_properties = main_properties + hessian_properties + dummy_properties
        self.all_properties = all_properties

        self.atomsdata = spk.data.AtomsData(dbpath, available_properties=all_properties)

        # The main file parser is always needed
        self.main_parser = OrcaMainFileParser(properties=main_properties + ["atoms"])

        if len(hessian_properties) > 0:
            self.hessian_parser = OrcaHessianFileParser(properties=hessian_properties)
        else:
            self.hessian_parser = None

        # Set up filter dictionary to e.g. remove numerically unstable solvent computations
        self.filter = filter

        # If requested, mask Q charges introduced by Orca
        self.mask_charges = mask_charges

    def parse_data(self, data_files, buffer_size=10):
        """
        Reads in a list of ORCA output files, extracts the data, performs reformatting and
        then stores structures and properties to an ASE database.

        Args:
            data_files (list): List of the paths to the ORCA output files.
            buffer_size (int, optional): Collects a certain number of molecules before writing them to
                                         the database.
        """
        atom_buffer = []
        property_buffer = []

        for file in tqdm(sorted(data_files), ncols=100):

            if os.path.exists(file):

                atoms, properties = self._parse_molecule(file)

                if properties is not None:
                    # Filter properties for problematic values
                    if self.filter is not None:
                        filtered = False
                        for p in self.filter:
                            if np.linalg.norm(properties[p]) > self.filter[p]:
                                filtered = True
                                logging.info(f"Filtered output {file} due to {p}")
                        if not filtered:
                            atom_buffer.append(atoms)
                            property_buffer.append(properties)
                    else:
                        atom_buffer.append(atoms)
                        property_buffer.append(properties)

                    if len(atom_buffer) >= buffer_size:
                        self.atomsdata.add_systems(atom_buffer, property_buffer)
                        atom_buffer = []
                        property_buffer = []

        # Collect leftovers
        if len(atom_buffer) > 0:
            self.atomsdata.add_systems(atom_buffer, property_buffer)

        metadata = {}
        self.atomsdata.set_metadata(metadata)

    def _parse_molecule(self, datafile):
        """
        Parser for a single molecule. This routine first checks for convergence
        of the calculation and then extracts all properties as instructed in the
        individual parsers.

        Args:
            datafile (str): Path to the datafile.

        Returns:
            (ase.Atoms, dict):
                atoms:
                    ase.Atoms object holding atom types and positions.
                properties:
                    Dictionary containing all extracted properties. Naming conventions are the
                    same as used in :obj:`schnetpack.Properties`.
        """
        # check if computation converged
        if not self._check_convergence(datafile):
            return None, None

        # Get main properties
        self.main_parser.parse_file(datafile)
        main_properties = self.main_parser.get_parsed()

        if self.mask_charges:
            main_properties = self._mask_charges(main_properties)

        atoms = None
        properties = {}
        for p in main_properties:
            if main_properties[p] is None:
                print("Error parsing {:s}".format(p))
                return None, None
            elif p == "atoms":
                atypes, coords = main_properties[p]
                atoms = Atoms(atypes, coords)
            else:
                properties[p] = main_properties[p].astype(np.float32)

        if self.hessian_parser is not None:
            hessian_file = (
                os.path.splitext(datafile)[0] + self.file_extensions["hessian"]
            )
            if not os.path.exists(hessian_file):
                print("Could not open Hessian file {:s}".format(hessian_file))
                return atoms, None
            else:
                self.hessian_parser.parse_file(hessian_file)
                hessian_properties = self.hessian_parser.get_parsed()

                for p in hessian_properties:
                    if p is None:
                        return atoms, None
                    elif p == Properties.dipole_derivatives:
                        properties[p] = format_dipole_derivatives(hessian_properties[p])
                    elif p == Properties.polarizability_derivatives:
                        properties[p] = format_polarizability_derivatives(
                            hessian_properties[p]
                        )
                    else:
                        properties[p] = hessian_properties[p]

        return atoms, properties

    @staticmethod
    def _check_convergence(datafile):
        """
        Check whether calculation has converged by searching for the
        typical ORCA closing statement.

        Args:
            datafile (str): Path to the output file.

        Returns:
            bool: Indicator if computation finished successfully.
        """
        flag = open(datafile).readlines()[-2].strip()

        if not flag == "****ORCA TERMINATED NORMALLY****":
            return False
        else:
            return True

    def _mask_charges(self, main_properties):
        """
        Remove the external charges Q introduced in orca input file. This is
        only necessary, if the charges are given in the input file. This in
        turn is necessary to get the right shielding tensors, as Orca is buggy
        in this case. All other properties need to be taken from a computation
        with external charges given in a file, since external charges in the
        input file are added to the potential energy and the dipole moment...

        Args:
            main_properties (dict): main property dictionary generated by the
                                    :obj:`OrcaParser`.

        Returns:
            dict: Updated property dictionary with the structures, atom types
                  and atomistic properties masked.
        """
        n_atoms = np.sum(main_properties["atoms"][0] != "Q")

        for p in main_properties:
            if p in self.atomistic:
                if p == "atoms":
                    main_properties[p][0] = main_properties[p][0][:n_atoms]
                    main_properties[p][1] = main_properties[p][1][:n_atoms]
                else:
                    main_properties[p] = main_properties[p][:n_atoms]

        return main_properties


def format_dipole_derivatives(property):
    """
    Reshape the extracted dipole derivatives to the correct
    format. Format is Natoms x (dx dy dz) x (property x y z)

    Args:
        property (numpy.array):

    Returns:
        numpy.array: Reshaped array
    """
    N, _ = property.shape
    N = N // 3
    property = property.reshape(N, 3, 3)
    return property


def format_polarizability_derivatives(property):
    """
    Reshape the extracted polarizability derivatives to the correct
    format. Format is Natoms x (dx dy dz) x (property Tensor)

    Args:
        property (numpy.array):

    Returns:
        numpy.array: Reshaped array
    """
    N, _ = property.shape
    N = N // 3
    property = property.reshape(N, 3, 6)
    triu_idx = np.triu_indices(3)
    reshaped = np.zeros((N, 3, 3, 3))
    reshaped[:, :, triu_idx[0], triu_idx[1]] = property
    reshaped[:, :, triu_idx[1], triu_idx[0]] = property
    return reshaped


class OrcaOutputParser:
    """
    Basic ORCA output parser class. Parses an Orca output file according to the parsers specified in the 'parsers'
    dictionary. Parsed data is stored in an dictionary, using the same keys as the parsers. If a list of formatters is
    provided to a parser, a list of the parsed entries is stored in the output dictionary.

    Args:
        parsers (dict[str->callable]): dictionary of :obj:`OrcaPropertyParser`,
                                       each with their own :obj:`OrcaFormatter`.
    """

    def __init__(self, parsers):
        self.parsers = parsers
        self.parsed = None

    def parse_file(self, path):
        """
        Open the file and iterate over its lines, applying all parsers. In the end, all data is collected in a
        dictionary.

        Args:
            path (str): path to Orca output file.
        """
        # Reset for new file
        for parser in self.parsers:
            self.parsers[parser].reset()

        with open(path, "r") as f:
            for line in f:
                for parser in self.parsers:
                    self.parsers[parser].parse_line(line)

        self.parsed = {}

        for parser in self.parsers:
            self.parsed[parser] = self.parsers[parser].get_parsed()

    def get_parsed(self):
        """
        Auxiliary routine to collect the data from the parser.

        Returns:
            dict[str->list]: Dictionary of data entries according to parser keys.
        """
        return self.parsed


class OrcaFormatter:
    """
    Format raw ORCA data collected by an :obj:`OrcaPropertyParser`. Behavior is determined by the datatype option.
    This is e.g. used to extract the correct gradient values from the associated text block taken from the ORCA
    output and convert it to a properly formatted force array.

    Args:
        position (int): Position to start formatting. If no stop is provided returns only value at position, otherwise
                        all values between position and stop are returned. (Only used for `vector` mode)
        stop (int, optional): Stop value for range. (Only used for `vector` mode)
        datatype (str, optional): Change formatting behavior. The possible options are:

                                   - `vector`:
                                        Formats data between position and stop argument, if provided
                                        converting it to the type given in the converter.
                                   - `matrix`:
                                        Formats collected matrix data into the shape of a square, symmetric
                                        numpy.ndarray. Ignores other options.
        converter (type, optional): Convert data to type. (Only used for 'vector' mode)
        default (float, optional): Default value to be returned if nothing is parsed (e.g. 1.0 for vacuum in case of
                         dielectric constant.
        skip_first (int, optional): If not None, skip the first N lines (default=None).
    """

    def __init__(
        self,
        position,
        stop=None,
        datatype="vector",
        converter=float,
        skip_first=None,
        unit=None,
        default=None,
    ):
        self.position = position
        self.stop = stop
        self.datatype = datatype
        self.converter = converter
        self.matrix_dim = None
        self.skip_first = skip_first
        self.unit = unit
        self.default = default

    def format(self, parsed):
        """
        Format the raw parsed data according to the given instructions.

        Args:
            parsed (list): List of raw parsed lines taken from the output file.

        Returns:
            numpy.array: Formatted numpy array holding the processed properties.
        """
        if parsed is None:
            if self.default is not None:
                return np.array([self.default])
            else:
                return None
        else:
            if self.skip_first is not None:
                parsed = parsed[self.skip_first :]

        if len(parsed) == 0:
            return None

        elif self.datatype == "vector":
            formatted = self._format_vector(parsed)
        elif self.datatype == "matrix":
            formatted = self._format_matrix(parsed)
        elif self.datatype == "shielding":
            formatted = self._format_shielding(parsed)
        else:
            raise NotImplementedError(
                "Unrecognized data type {:s}".format(self.datatype)
            )

        if self.unit is not None:
            formatted *= self.unit

        return formatted

    def _format_vector(self, parsed):
        """
        Take numerical entries in a line and collect them into an array.
        It is possible to extract only certain slices of an array.
        Although the basic quantity is called `vector`, this also is used to
        extract matrices, such as forces. The special matrix formatter deals
        with cases where ORCA stores special matrices, such as Hamiltionians
        and Hessians. In this case it introduces line breaks after six entries,
        leading this routine to fail.

        Args:
            parsed (list): List of raw parsed lines taken from the output file.

        Returns:
            numpy.array: Formatted numpy array holding the processed properties.
        """
        vector = []
        for line in parsed:
            line = line.split()
            if self.stop is None:
                vector.append(self.converter(line[self.position]))
            else:
                vector.append(
                    [self.converter(x) for x in line[self.position : self.stop]]
                )

        vector = np.array(vector)

        # Remove trailing dimension if only one line is read (for dipole moment)
        if vector.shape[0] == 1 and vector.size != 1:
            vector = vector[0]

        return vector

    def _format_matrix(self, parsed):
        """
        Format raw extracted matrices. Unlike the vector formatter, this routine
        deals with cases where ORCA stores special matrices, such as Hamiltionians
        and Hessians. In this case it introduces line breaks after six entries,
        which have to be reformatted.

        Args:
            parsed (list): List of raw parsed lines taken from the output file.

        Returns:
            numpy.array: Formatted numpy array holding the processed properties.
        """
        n_entries = len(parsed[1].split())

        # Get matrix dimensions
        for line in parsed[1:]:
            line = line.split()
            if len(line) != n_entries:
                self.matrix_dim = int(line[0]) + 1

        subdata = [
            parsed[i : i + self.matrix_dim + 1]
            for i in range(0, len(parsed), self.matrix_dim + 1)
        ]

        matrix = [[] for _ in range(self.matrix_dim)]

        for block in subdata:
            for i, entry in enumerate(block[1:]):
                matrix[i] += [float(x) for x in entry.split()[1:]]

        matrix = np.array(matrix)
        return matrix

    @staticmethod
    def _format_shielding(parsed):
        """
        Format the raw shielding tensors taken from the ORCA output.

        Args:
            parsed (list): List of raw parsed lines taken from the output file.

        Returns:
            numpy.array: Formatted numpy array holding the processed properties.
        """
        shielding = []
        current_shielding = []
        parse = False
        for line in parsed:
            if line.startswith("Total shielding tensor (ppm):"):
                parse = True
            elif parse:
                if line.startswith("Diagonalized sT*s matrix:"):
                    shielding.append(current_shielding)
                    current_shielding = []
                    parse = False
                else:
                    current_shielding.append([float(x) for x in line.split()])
            else:
                continue

        shielding = np.array(shielding)
        return shielding


class OrcaPropertyParser:
    """
    Basic property parser for ORCA output files. Takes a start flag and a stop flag/list of stop flags and collects
    the data entries in between. If a :obj:`OrcaFormatter` is provided, the data is formatted accordingly upon
    retrieval. Operates in a line-wise fashion.

    Args:
        start (str): begins to collect data starting from this string
        stop (str/list(str)): stops data collection if any of these strings is encounteres
        formatters (OrcaFormatter): OrcaFormatter to convert collected data
    """

    def __init__(self, start, stop, formatters=None):
        self.start = start
        self.stop = stop
        self.formatters = formatters

        self.read = False
        self.parsed = None

    def parse_line(self, line):
        """
        Parses a line in the output file and updates the main container.

        Args:
            line (str): line of Orca output file
        """
        line = line.strip()
        if line.startswith("---------") or len(line) == 0:
            pass
        # if line.startswith("*********") or len(line) == 0:
        #     pass
        elif line.startswith(self.start):
            # Avoid double reading and restart for multiple files and repeated instances of data.
            self.parsed = []
            self.read = True
            # For single line output
            if self.stop is None:
                self.parsed.append(line)
                self.read = False
        elif self.read:
            # Check for stops
            if isinstance(self.stop, list):
                for stop in self.stop:
                    if self.read and line.startswith(stop):
                        self.read = False
                if self.read:
                    self.parsed.append(line)
            else:
                if line.startswith(self.stop):
                    self.read = False
                else:
                    self.parsed.append(line)

    def get_parsed(self):
        """
        Returns data, if formatters are specified in the corresponding format.

        Returns:
            numpy.array: Formatted data.
        """
        if self.formatters is None:
            return self.parsed
        elif hasattr(self.formatters, "__iter__"):
            return [formatter.format(self.parsed) for formatter in self.formatters]
        else:
            return self.formatters.format(self.parsed)

    def reset(self):
        """
        Reset state of the parser.
        """
        self.read = False
        self.parsed = None


class OrcaMainFileParser(OrcaOutputParser):
    """
    Predefined :obj:`OrcaOutputParser` for extracting data from the main ORCA output file.
    Can read and format structure and atom types, as well as the energies, forces, dipole moments,
    polarizabilities and nuclear shielding tensors.
    """

    properties = [
        "atoms",
        Properties.forces,
        Properties.energy,
        Properties.dipole_moment,
        Properties.polarizability,
        Properties.shielding,
    ]

    starts = {
        "atoms": "CARTESIAN COORDINATES (ANGSTROEM)",
        Properties.forces: "CARTESIAN GRADIENT",
        Properties.energy: "FINAL SINGLE POINT ENERGY",
        Properties.dipole_moment: "Total Dipole Moment",
        Properties.polarizability: "The raw cartesian tensor (atomic units):",
        Properties.shielding: "CHEMICAL SHIFTS",
    }

    stops = {
        "atoms": "CARTESIAN COORDINATES (A.U.)",
        Properties.forces: "Difference to translation invariance",
        Properties.energy: None,
        Properties.dipole_moment: None,
        Properties.polarizability: "diagonalized tensor:",
        Properties.shielding: "CHEMICAL SHIELDING SUMMARY",
    }

    formatters = {
        "atoms": (
            OrcaFormatter(0, converter=str),
            OrcaFormatter(1, stop=4, unit=1.0 / units.Bohr),
        ),
        Properties.energy: OrcaFormatter(4),
        Properties.forces: OrcaFormatter(3, stop=6, unit=-1.0),
        Properties.dipole_moment: OrcaFormatter(4, stop=7),
        Properties.polarizability: OrcaFormatter(0, stop=4),
        Properties.shielding: OrcaFormatter(None, datatype="shielding", unit=ppm2au),
    }

    def __init__(self, properties=None):

        if properties is None:
            to_parse = self.properties
        else:
            to_parse = []
            for p in properties:
                if p not in self.properties:
                    print("Cannot parse property {:s}".format(p))
                else:
                    to_parse.append(p)

        parsers = {
            p: OrcaPropertyParser(
                self.starts[p], self.stops[p], formatters=self.formatters[p]
            )
            for p in to_parse
        }

        super(OrcaMainFileParser, self).__init__(parsers)


class OrcaHessianFileParser(OrcaMainFileParser):
    """
    Predefined :obj:`OrcaOutputParser` for extracting data from the hessian output file written
    by ORCA if some higher order derivatives are requested. Can read and format Hessians, as well as
    Cartesian dipole moment and polarizability derivatives.
    """

    properties = [
        Properties.hessian,
        Properties.dipole_derivatives,
        Properties.polarizability_derivatives,
    ]

    starts = {
        Properties.hessian: "$hessian",
        Properties.dipole_derivatives: "$dipole_derivatives",
        Properties.polarizability_derivatives: "$polarizability_derivatives",
    }

    stops = {
        Properties.hessian: "$vibrational_frequencies",
        Properties.dipole_derivatives: "#",
        Properties.polarizability_derivatives: "#",
    }

    formatters = {
        Properties.hessian: OrcaFormatter(None, datatype="matrix", skip_first=1),
        Properties.dipole_derivatives: OrcaFormatter(0, stop=4, skip_first=1),
        Properties.polarizability_derivatives: OrcaFormatter(0, stop=6, skip_first=1),
    }

    def __init__(self, properties=None):
        super(OrcaHessianFileParser, self).__init__(properties)



r"""
Classes wrapping various standard benchmark datasets.
"""

from schnetpack import AtomsData, get_center_of_mass
from schnetpack.data.atoms import logger
from schnetpack.environment import SimpleEnvironmentProvider


class DownloadableAtomsData(AtomsData):
    """
    Base class for online datasets that can be automatically downloaded.

        Args:
            dbpath (str): path to directory containing database.
            subset (list, optional): Deprecated! Do not use! Subsets are created with
                AtomsDataSubset class.
            available_properties (list, optional): complete set of physical properties
                that are contained in the database.
            load_only (list, optional): reduced set of properties to be loaded
            units (list, optional): definition of units for all available properties
            environment_provider (spk.environment.BaseEnvironmentProvider): define how
                neighborhood is calculated
                (default=spk.environment.SimpleEnvironmentProvider).
            collect_triples (bool, optional): Set to True if angular features are
            needed.
            centering_function (callable or None): Function for calculating center of
                molecule (center of mass/geometry/...). Center will be subtracted from
                positions.
            download (bool): If true, automatically download dataset
                if it does not exist.
    """

    def __init__(
        self,
        dbpath,
        subset=None,
        load_only=None,
        available_properties=None,
        units=None,
        environment_provider=SimpleEnvironmentProvider(),
        collect_triples=False,
        centering_function=get_center_of_mass,
        download=False,
    ):

        super(DownloadableAtomsData, self).__init__(
            dbpath=dbpath,
            subset=subset,
            available_properties=available_properties,
            load_only=load_only,
            units=units,
            environment_provider=environment_provider,
            collect_triples=collect_triples,
            centering_function=centering_function,
        )
        if download:
            self.download()

    def download(self):
        """
        Wrapper function for the download method.
        """
        if os.path.exists(self.dbpath):
            logger.info(
                "The dataset has already been downloaded and stored "
                "at {}".format(self.dbpath)
            )
        else:
            logger.info("Starting download")
            folder = os.path.dirname(os.path.abspath(self.dbpath))
            if not os.path.exists(folder):
                os.makedirs(folder)
            self._download()

    def _download(self):
        """
        To be implemented in deriving classes.
        """
        raise NotImplementedError


from .ani1 import *
from .iso17 import *
from .matproj import *
from .omdb import *
from .md17 import *
from .qm9 import *
from .extxyz import ExtXYZ



"""
Neighbor lists are used to obtain the indices of neighbors surrounding an atom
for the :obj:`schnetpack.md.calculators.SchnetPackCalculator`.
Currently only a primitive version of a neighbor list is implemented, which cannot deal
with periodic boundary conditions and does not possess optimal scaling for large systems.
"""

import torch
import schnetpack


class MDNeighborList:
    """
    Basic neighbor list template for molecular dynamics simulations required for the calculator. This is used to obtain
    the indices of the atoms adjacent to a central atom and e.g. used to compute the molecular interactions.
    The neighbor mask is zero for interactions which should not be counted and one otherwise.

    Args:
        cutoff (float): Cutoff radius used for neighbor list construction.
        shell (float): Buffer region around the cutoff radius. A new neighbor list is only constructed if an atom
                       moves farther than this shell. (Or if the simulation cell changes.)
    """

    def __init__(self, cutoff, shell=None, device=None):
        self.device = device

        # Check cutoff and shell, as well as possible conventions
        self.cutoff = cutoff
        self.shell = shell

        if self.cutoff is not None:
            if self.shell is not None:
                self.cutoff_shell = self.cutoff + self.shell
            else:
                self.cutoff_shell = self.cutoff
        else:
            self.cutoff_shell = None

        # Init basic containers
        self.neighbor_list = None
        self.neighbor_mask = None
        self.offsets = None
        self.max_neighbors = None

        # Store last positions and cells for determining whether the neighbor list needs to be recomputed
        self.last_positions = None
        self.last_cells = None

    def get_neighbors(self, system):
        """
        Convenience routine to obtain the neighbor list and neighbor mask in one step.

        Returns:
            tuple: Contains the neighbor list and neighbor mask tensors.
        """
        if self._update_required(system) or self.neighbor_list is None:
            self._construct_neighbor_list(system)

        return self.neighbor_list, self.neighbor_mask, self.offsets

    def _construct_neighbor_list(self, system):
        """
        Instructions to construct the neighbor list. Needs to be defined and has to populate the neighbor_list
        and neighbor_mask tensors.

        Both, neighbor_list and neighbor_mask, should be a torch.Tensor with the dimensions:
            n_replicas x n_molecules x n_atoms x n_neighbors
        """
        raise NotImplementedError

    def _update_required(self, system):
        """
        Function to determine whether the neighbor list should be recomputed for the system. This could e.g. be
        based on the maximum distance all atoms moved, etc. The return value should be True if it needs to be
        recomputed and False otherwise.

        Returns:
            bool: Indicator whether it is necessary to compute a new neighbor list or not.
        """
        # Compute the neighbor list if these two are not set
        if self.last_positions is None or self.last_cells is None:
            return True

        # Check if cell has changed
        if not torch.allclose(self.last_cells, system.cells):
            return True

        # Check if atoms have moved out of the boundary
        max_displacement = torch.max(
            torch.norm(system.positions - self.last_positions, 2, 3)
        ).detach()
        if max_displacement >= self.shell:
            return True

        return False


class SimpleNeighborList(MDNeighborList):
    """
    Basic implementation of a neighbor list. Simply enumerates the neighbors of all atoms in the molecule after
    eliminating self interactions. This work fine for small to medium sized systems, but should not be used for
    extended molecules, etc. The cutoff fulfills no purpose in this basic implementation. This neighbor list should
    never be used in combination with periodic boundary conditions.

    Args:
        system (object): System class containing all molecules and their replicas.
        cutoff (float): Cutoff radius used for neighbor list construction, not used in the present implementation.
    """

    def __init__(self, cutoff=None, shell=None, device=None):
        super(SimpleNeighborList, self).__init__(cutoff, shell, device=device)
        self.n_replicas = None

    def _construct_neighbor_list(self, system):
        """
        Sets up a basic neighbor list, neighbor mask and offset array. The first two are torch.Tensor objects of the
        dimension: n_replicas x n_molecules x n_atoms x n_neighbors. The offsets have the dimension
        n_replicas x n_molecules x n_atoms x n_neighbors x 3.
        This neighbor list simply enumerates all neighbors (neighbor_list) or mask nonsensical entries due to either
        different cutoff radii or zero-padding arising from molecules of different size (neighbor_mask).
        """
        # Set the maximum neighbors to include all interactions
        self.max_neighbors = system.max_n_atoms - 1
        self.n_replicas = system.n_replicas

        # Construct basic, unmasked tile
        basic_tile = torch.arange(system.max_n_atoms, device=system.device)[
            None, :
        ].repeat(system.max_n_atoms, 1)
        # Remove self interactions
        diagonal_mask = torch.eye(system.max_n_atoms, device=system.device)
        basic_tile = basic_tile[diagonal_mask != 1].view(
            system.max_n_atoms, system.max_n_atoms - 1
        )

        # Tile neighbor lists and mask to get replica and molecule dimensions
        neighbors_list = basic_tile[None, None, :, :].repeat(
            system.n_replicas, system.n_molecules, 1, 1
        )

        # Construct the neighbor mask as an outer product of the atom masks, where self interactions are removed
        self.neighbor_mask = (
            system.atom_masks.transpose(2, 3)[..., 1:] * system.atom_masks
        )

        # Construct cell offsets
        if system.cells is not None:
            self.offsets = torch.zeros(
                system.n_replicas,
                system.n_molecules,
                system.max_n_atoms,
                self.max_neighbors,
                3,
                device=system.device,
            )
        else:
            self.offsets = None

        # Multiply neighbor list by mask to remove superfluous entries
        self.neighbor_list = neighbors_list * self.neighbor_mask.long()

    def _update_required(self, system):
        """
        Since all interatomic distances are computed by default, the neighbor list never has to be updated.

        Returns:
            bool: Indicator whether it is necessary to compute a new neighbor list or not.
        """
        if self.n_replicas is not None and self.n_replicas == system.n_replicas:
            return False
        else:
            return True


class EnvironmentProviderNeighborList(MDNeighborList):
    """
    Basic neighbor list class to be used with the environment providers with SchNetPack. The corresponding provider
    needs to be set in the `_set_environment_provider` function. Since this currently operates sequentially, it will
    provide suboptimal performance for systems with many replicas and/or molecules.

    Args:
        cutoff (float): Cutoff radius used for neighbor list construction.
        shell (float): Buffer region around the cutoff radius. A new neighbor list is only constructed if an atom
                       moves farther than this shell. (Or if the simulation cell changes.)
    """

    def __init__(self, cutoff, shell=1.0, device=None, use_internal_units=True):
        super(EnvironmentProviderNeighborList, self).__init__(
            cutoff, shell, device=device
        )
        self.use_internal_units = use_internal_units
        self.provider_cutoff = self._get_provider_cutoff()

        # Setup the environment provider
        self._environment_provider = None
        self._set_environment_provider()

    def _get_provider_cutoff(self):
        if self.use_internal_units:
            provider_cutoff = self.cutoff_shell
        else:
            provider_cutoff = self.cutoff_shell / schnetpack.md.MDUnits.angs2internal
        return provider_cutoff

    def _set_environment_provider(self):
        """
        This function is intended to set the environment provider in neighbor lists based on this class.
        """
        raise NotImplementedError

    def _construct_neighbor_list(self, system):
        """
        Construct the neighbor list using an environment provider. Since all providers are based on ASE atoms objects,
        these objects are first extracted from the system. Then the neighbor lists ae constructed sequentially and
        reconverted into the format required for the calculators. In addition, the old cells and positons are
        stored to check if updates of the neighbor list are necessary.
        """
        atoms = system.get_ase_atoms(internal_units=self.use_internal_units)

        neighbor_idx = []
        offsets = []
        max_neighbors = 0

        for mol in atoms:
            nbh_idx, offset = self._environment_provider.get_environment(mol)
            neighbor_idx.append(nbh_idx)
            offsets.append(offset)
            max_neighbors = max(max_neighbors, nbh_idx.shape[1])

        self.neighbor_list = -torch.ones(
            system.n_replicas,
            system.n_molecules,
            system.max_n_atoms,
            max_neighbors,
            device=system.device,
        ).long()

        self.offsets = torch.zeros(
            system.n_replicas,
            system.n_molecules,
            system.max_n_atoms,
            max_neighbors,
            3,
            device=system.device,
        )

        count = 0
        for r_idx in range(system.n_replicas):
            for m_idx in range(system.n_molecules):
                n_atoms = system.n_atoms[m_idx]
                n_nbh = neighbor_idx[count].shape[1]
                self.neighbor_list[r_idx, m_idx, :n_atoms, :n_nbh] = torch.from_numpy(
                    neighbor_idx[count]
                )

                if system.cells is not None:
                    self.offsets[r_idx, m_idx, :n_atoms, :n_nbh, :] = torch.from_numpy(
                        offsets[count]
                    )
                else:
                    self.offsets = None

                count += 1

        self.max_neighbors = max_neighbors
        self.neighbor_mask = torch.zeros_like(self.neighbor_list)
        self.neighbor_mask[self.neighbor_list >= 0] = 1.0

        # Mask away -1 indices for invalid atoms, since they do not work with torch.gather
        self.neighbor_list = self.neighbor_list * self.neighbor_mask.long()

        # Since the list was recomputed, update old cells and positions
        self.last_positions = system.positions.clone().detach()

        if system.cells is not None:
            self.last_cells = system.cells.clone().detach()


class ASENeighborList(EnvironmentProviderNeighborList):
    """
    Neighbor list based on the schnetpack.utils.environment.AseEnvironmentProvider. This can deal with periodic
    boundary conditions and general unit cells. However, the provider runs on CPU only and will only provide
    significant performance gains over the torch based one for very large systems.
    The ASE neighbor_list internally uses a minimum bin size of 3A, hence positions and cells need to be converted
    to A before passing them to the neighbor list to avoid performance issues.

    Args:
        cutoff (float): Cutoff radius used for neighbor list construction.
        shell (float): Buffer region around the cutoff radius. A new neighbor list is only constructed if an atom
                       moves farther than this shell. (Or if the simulation cell changes.)
    """

    def __init__(self, cutoff, shell, device=None):
        super(ASENeighborList, self).__init__(
            cutoff=cutoff, shell=shell, device=device, use_internal_units=False
        )

    def _set_environment_provider(self):
        """
        Set the environment provider.
        """
        self._environment_provider = schnetpack.environment.AseEnvironmentProvider(
            self.provider_cutoff
        )


class TorchNeighborList(EnvironmentProviderNeighborList):
    """
    Neighbor list based on the schnetpack.utils.environment.TorchEnvironmentProvider. For moderately sized systems
    with cells/periodic boundary conditions this should have a good performance.

    Args:
        cutoff (float): Cutoff radius used for neighbor list construction.
        shell (float): Buffer region around the cutoff radius. A new neighbor list is only constructed if an atom
                       moves farther than this shell. (Or if the simulation cell changes.)
        device (torch.device): Device used when computing the neighbor list.
    """

    def __init__(self, cutoff, shell, device=torch.device("cpu")):
        super(TorchNeighborList, self).__init__(
            cutoff=cutoff, shell=shell, device=device
        )

    def _set_environment_provider(self):
        """
        Set the environment provider.
        """
        self._environment_provider = schnetpack.environment.TorchEnvironmentProvider(
            self.cutoff_shell, self.device
        )


class DualNeighborList:
    def __init__(
        self, cutoff_short, cutoff_long, neighbor_list, shell=1.0, device=None
    ):
        self.neighbor_list_short = neighbor_list(cutoff_short, shell, device=device)
        self.neighbor_list_long = neighbor_list(cutoff_long, shell, device=device)

    def get_neighbors(self, system):
        neighbors, neighbor_mask, offsets = self.neighbor_list_short.get_neighbors(
            system
        )
        return neighbors, neighbor_mask, offsets

    def get_neighbors_lr(self, system):
        (
            neighbors_long,
            neighbor_mask_long,
            offsets_long,
        ) = self.neighbor_list_long.get_neighbors(system)
        return neighbors_long, neighbor_mask_long, offsets_long

    @property
    def max_neighbors(self):
        return self.neighbor_list_short.max_neighbors

    @property
    def max_neighbors_lr(self):
        return self.neighbor_list_long.max_neighbors



import math

import numpy as np
import torch
from ase.neighborlist import neighbor_list

__all__ = [
    "BaseEnvironmentProvider",
    "SimpleEnvironmentProvider",
    "AseEnvironmentProvider",
    "TorchEnvironmentProvider",
]


class BaseEnvironmentProvider:
    """
    Environment Providers are supposed to collect neighboring atoms within
    local, atom-centered environments. All environment providers should inherit
    from this class.

    """

    def get_environment(self, atoms):
        """
        Returns the neighbor indices and offsets

        Args:
            atoms (ase.Atoms): atomistic system

        Returns:
            neighborhood_idx (np.ndarray): indices of the neighbors with shape
                n_atoms x n_max_neighbors
            offset (np.ndarray): offset in lattice coordinates for periodic
                systems (otherwise zero matrix) of shape
                n_atoms x n_max_neighbors x 3

        """

        raise NotImplementedError


class SimpleEnvironmentProvider(BaseEnvironmentProvider):
    """
    A simple environment provider for small molecules where all atoms are each
    other's neighbors. It calculates full distance matrices and does not
    support cutoffs or periodic boundary conditions.
    """

    def get_environment(self, atoms, grid=None):
        n_atoms = atoms.get_global_number_of_atoms()

        if n_atoms == 1:
            neighborhood_idx = -np.ones((1, 1), dtype=np.float32)
            offsets = np.zeros((n_atoms, 1, 3), dtype=np.float32)
        else:
            neighborhood_idx = np.tile(
                np.arange(n_atoms, dtype=np.float32)[np.newaxis], (n_atoms, 1)
            )

            neighborhood_idx = neighborhood_idx[
                ~np.eye(n_atoms, dtype=np.bool)
            ].reshape(n_atoms, n_atoms - 1)

            if grid is not None:
                n_grid = grid.shape[0]
                neighborhood_idx = np.hstack([neighborhood_idx, -np.ones((n_atoms, 1))])
                grid_nbh = np.tile(
                    np.arange(n_atoms, dtype=np.float32)[np.newaxis], (n_grid, 1)
                )
                neighborhood_idx = np.vstack([neighborhood_idx, grid_nbh])

            offsets = np.zeros(
                (neighborhood_idx.shape[0], neighborhood_idx.shape[1], 3),
                dtype=np.float32,
            )
        return neighborhood_idx, offsets


class AseEnvironmentProvider(BaseEnvironmentProvider):
    """
    Environment provider making use of ASE neighbor lists. Supports cutoffs
    and PBCs.

    """

    def __init__(self, cutoff):
        self.cutoff = cutoff

    def get_environment(self, atoms, grid=None):
        if grid is not None:
            raise NotImplementedError

        n_atoms = atoms.get_global_number_of_atoms()
        idx_i, idx_j, idx_S = neighbor_list(
            "ijS", atoms, self.cutoff, self_interaction=False
        )
        if idx_i.shape[0] > 0:
            uidx, n_nbh = np.unique(idx_i, return_counts=True)
            n_max_nbh = np.max(n_nbh)

            n_nbh = np.tile(n_nbh[:, np.newaxis], (1, n_max_nbh))
            nbh_range = np.tile(
                np.arange(n_max_nbh, dtype=np.int)[np.newaxis], (n_nbh.shape[0], 1)
            )

            mask = np.zeros((n_atoms, np.max(n_max_nbh)), dtype=np.bool)
            mask[uidx, :] = nbh_range < n_nbh
            neighborhood_idx = -np.ones((n_atoms, np.max(n_max_nbh)), dtype=np.float32)
            neighborhood_idx[mask] = idx_j

            offset = np.zeros((n_atoms, np.max(n_max_nbh), 3), dtype=np.float32)
            offset[mask] = idx_S
        else:
            neighborhood_idx = -np.ones((n_atoms, 1), dtype=np.float32)
            offset = np.zeros((n_atoms, 1, 3), dtype=np.float32)

        return neighborhood_idx, offset


class TorchEnvironmentProvider(BaseEnvironmentProvider):
    """
    Environment provider making use of neighbor lists as implemented in TorchAni
    (https://github.com/aiqm/torchani/blob/master/torchani/aev.py).
    Supports cutoffs, PBCs and can be performed on either CPU or GPU.

    """

    def __init__(self, cutoff, device):
        """
        Args:
            cutoff (float): the cutoff inside which atoms are considered pairs
            device (:class:`torch.device`): pass torch.device('cpu') or torch.device('cuda') to
                perform the calculation on a CPU or GPU, respectively.
        """
        self.cutoff = cutoff
        self.device = device

    def get_environment(self, atoms):

        species = torch.FloatTensor(atoms.numbers).to(self.device)
        coordinates = torch.FloatTensor(atoms.positions).to(self.device)
        pbc = torch.from_numpy(atoms.pbc.astype("uint8")).to(self.device)

        if not atoms.cell.any():
            cell = torch.eye(3, dtype=species.dtype).to(self.device)
        else:
            cell = torch.Tensor(atoms.cell).to(self.device)

        shifts = compute_shifts(cell=cell, pbc=pbc, cutoff=self.cutoff)

        # The returned indices are only one directional
        idx_i, idx_j, idx_S = neighbor_pairs(
            species == -1, coordinates, cell, shifts, self.cutoff
        )

        idx_i = idx_i.cpu().detach().numpy()
        idx_j = idx_j.cpu().detach().numpy()
        idx_S = idx_S.cpu().detach().numpy()

        # Create bidirectional id arrays, similar to what the ASE neighbor_list returns
        bi_idx_i = np.hstack((idx_i, idx_j))
        bi_idx_j = np.hstack((idx_j, idx_i))
        bi_idx_S = np.vstack((-idx_S, idx_S))

        n_atoms = atoms.get_global_number_of_atoms()
        if bi_idx_i.shape[0] > 0:
            uidx, n_nbh = np.unique(bi_idx_i, return_counts=True)
            n_max_nbh = np.max(n_nbh)

            n_nbh = np.tile(n_nbh[:, np.newaxis], (1, n_max_nbh))
            nbh_range = np.tile(
                np.arange(n_max_nbh, dtype=np.int)[np.newaxis], (n_nbh.shape[0], 1)
            )

            mask = np.zeros((n_atoms, np.max(n_max_nbh)), dtype=np.bool)
            mask[uidx, :] = nbh_range < n_nbh
            neighborhood_idx = -np.ones((n_atoms, np.max(n_max_nbh)), dtype=np.float32)
            offset = np.zeros((n_atoms, np.max(n_max_nbh), 3), dtype=np.float32)

            # Assign neighbors and offsets according to the indices in bi_idx_i, since in contrast
            # to the ASE provider the bidirectional arrays are no longer sorted.
            # TODO: There might be a more efficient way of doing this than a loop
            for idx in range(n_atoms):
                neighborhood_idx[idx, mask[idx]] = bi_idx_j[bi_idx_i == idx]
                offset[idx, mask[idx]] = bi_idx_S[bi_idx_i == idx]

        else:
            neighborhood_idx = -np.ones((n_atoms, 1), dtype=np.float32)
            offset = np.zeros((n_atoms, 1, 3), dtype=np.float32)

        return neighborhood_idx, offset


def compute_shifts(cell, pbc, cutoff):
    """Compute the shifts of unit cell along the given cell vectors to make it
    large enough to contain all pairs of neighbor atoms with PBC under
    consideration.
    Copyright 2018- Xiang Gao and other ANI developers
    (https://github.com/aiqm/torchani/blob/master/torchani/aev.py)

    Arguments:
        cell (:class:`torch.Tensor`): tensor of shape (3, 3) of the three
        vectors defining unit cell:
            tensor([[x1, y1, z1], [x2, y2, z2], [x3, y3, z3]])
        cutoff (float): the cutoff inside which atoms are considered pairs
        pbc (:class:`torch.Tensor`): boolean vector of size 3 storing
            if pbc is enabled for that direction.

    Returns:
        :class:`torch.Tensor`: long tensor of shifts. the center cell and
            symmetric cells are not included.
    """
    # type: (Tensor, Tensor, float) -> Tensor
    reciprocal_cell = cell.inverse().t()
    inv_distances = reciprocal_cell.norm(2, -1)
    num_repeats = torch.ceil(cutoff * inv_distances).to(torch.long)
    num_repeats = torch.where(pbc.bool(), num_repeats, torch.zeros_like(num_repeats))

    r1 = torch.arange(1, num_repeats[0] + 1, device=cell.device)
    r2 = torch.arange(1, num_repeats[1] + 1, device=cell.device)
    r3 = torch.arange(1, num_repeats[2] + 1, device=cell.device)
    o = torch.zeros(1, dtype=torch.long, device=cell.device)

    return torch.cat(
        [
            torch.cartesian_prod(r1, r2, r3),
            torch.cartesian_prod(r1, r2, o),
            torch.cartesian_prod(r1, r2, -r3),
            torch.cartesian_prod(r1, o, r3),
            torch.cartesian_prod(r1, o, o),
            torch.cartesian_prod(r1, o, -r3),
            torch.cartesian_prod(r1, -r2, r3),
            torch.cartesian_prod(r1, -r2, o),
            torch.cartesian_prod(r1, -r2, -r3),
            torch.cartesian_prod(o, r2, r3),
            torch.cartesian_prod(o, r2, o),
            torch.cartesian_prod(o, r2, -r3),
            torch.cartesian_prod(o, o, r3),
        ]
    )


def neighbor_pairs(padding_mask, coordinates, cell, shifts, cutoff):
    """Compute pairs of atoms that are neighbors
    Copyright 2018- Xiang Gao and other ANI developers
    (https://github.com/aiqm/torchani/blob/master/torchani/aev.py)

    Arguments:
        padding_mask (:class:`torch.Tensor`): boolean tensor of shape
            (molecules, atoms) for padding mask. 1 == is padding.
        coordinates (:class:`torch.Tensor`): tensor of shape
            (molecules, atoms, 3) for atom coordinates.
        cell (:class:`torch.Tensor`): tensor of shape (3, 3) of the three vectors
            defining unit cell: tensor([[x1, y1, z1], [x2, y2, z2], [x3, y3, z3]])
        cutoff (float): the cutoff inside which atoms are considered pairs
        shifts (:class:`torch.Tensor`): tensor of shape (?, 3) storing shifts
    """
    # type: (Tensor, Tensor, Tensor, Tensor, float) -> Tuple[Tensor, Tensor, Tensor, Tensor]

    coordinates = coordinates.detach()
    cell = cell.detach()
    num_atoms = padding_mask.shape[0]
    all_atoms = torch.arange(num_atoms, device=cell.device)

    # Step 2: center cell
    p1_center, p2_center = torch.combinations(all_atoms).unbind(-1)
    shifts_center = shifts.new_zeros(p1_center.shape[0], 3)

    # Step 3: cells with shifts
    # shape convention (shift index, molecule index, atom index, 3)
    num_shifts = shifts.shape[0]
    all_shifts = torch.arange(num_shifts, device=cell.device)
    shift_index, p1, p2 = torch.cartesian_prod(all_shifts, all_atoms, all_atoms).unbind(
        -1
    )
    shifts_outside = shifts.index_select(0, shift_index)

    # Step 4: combine results for all cells
    shifts_all = torch.cat([shifts_center, shifts_outside])
    p1_all = torch.cat([p1_center, p1])
    p2_all = torch.cat([p2_center, p2])

    shift_values = torch.mm(shifts_all.to(cell.dtype), cell)

    # step 5, compute distances, and find all pairs within cutoff
    distances = (coordinates[p1_all] - coordinates[p2_all] + shift_values).norm(2, -1)

    padding_mask = (padding_mask[p1_all]) | (padding_mask[p2_all])
    distances.masked_fill_(padding_mask, math.inf)
    in_cutoff = torch.nonzero(distances < cutoff, as_tuple=False)
    pair_index = in_cutoff.squeeze()
    atom_index1 = p1_all[pair_index]
    atom_index2 = p2_all[pair_index]
    shifts = shifts_all.index_select(0, pair_index)
    return atom_index1, atom_index2, shifts


def collect_atom_triples(nbh_idx):
    """
    Collect all valid triples of atoms by rearranging neighbor indices obtained
    from an environment provider.

    Args:
        nbh_idx (numpy.ndarray): neighbor indices

    Returns:
        nbh_idx_j, nbh_idx_k (numpy.ndarray): triple indices
        offset_idx_j, offset_idx_k (numpy.ndarray): offset indices for PBC

    """
    natoms, nneigh = nbh_idx.shape

    # Construct possible permutations
    nbh_idx_j = np.tile(nbh_idx, nneigh)
    nbh_idx_k = np.repeat(nbh_idx, nneigh).reshape((natoms, -1))

    # Remove same interactions and non unique pairs
    triu_idx_row, triu_idx_col = np.triu_indices(nneigh, k=1)
    triu_idx_flat = triu_idx_row * nneigh + triu_idx_col
    nbh_idx_j = nbh_idx_j[:, triu_idx_flat]
    nbh_idx_k = nbh_idx_k[:, triu_idx_flat]

    # Keep track of periodic images
    offset_idx = np.tile(np.arange(nneigh), (natoms, 1))

    # Construct indices for pairs of offsets
    offset_idx_j = np.tile(offset_idx, nneigh)
    offset_idx_k = np.repeat(offset_idx, nneigh).reshape((natoms, -1))

    # Remove non-unique pairs and diagonal
    offset_idx_j = offset_idx_j[:, triu_idx_flat]
    offset_idx_k = offset_idx_k[:, triu_idx_flat]

    return nbh_idx_j, nbh_idx_k, offset_idx_j, offset_idx_k


def collect_atom_triples_batch(neighbors, neighbor_mask):
    """
    Batch/torch version for collecting atom triples, offset indices and the corresponding mask directly from a
    batch of neighbor indices and their corresponding mask. This is e.g. used in the
    schnetpack.md.calculators.SchnetPackCalculator class to generate extended inputs for Behler type symmetry
    functions.
    Args:
        neighbors (torch.LongTensor): (n_batch x n_atoms x n_neighbors) tensor holding the indices of all
                                      neighbor atoms (e.g. from NeighborList or EnvironmentProvider).
        neighbor_mask (torch.LongTensor): (n_batch x n_atoms x n_neighbors) binary tensor indicating non-existent
                                          atoms due to padding.
    Returns:
        torch.LongTensor: (n_batch x n_atoms x n_triples) indices of the first neighbor in all triples.
        torch.LongTensor: (n_batch x n_atoms x n_triples) indices of the second neighbor in all triples.
        torch.LongTensor: (n_batch x n_atoms x n_triples) first neighbor offset indices for PBC.
        torch.LongTensor: (n_batch x n_atoms x n_triples) second neighbor offset indices for PBC.
        torch.LongTensor: (n_batch x n_atoms x n_triples) mask indicating all invalid pairs due to padding.
    """
    B, A, N = neighbors.shape

    # Generate indices of all possible unique pairs
    idx_k, idx_j = torch.combinations(
        torch.arange(N, device=neighbors.device).long(), r=2, with_replacement=False
    ).unbind(1)

    # Generate triple indices
    nbh_idx_j = neighbors[:, :, idx_j]
    nhb_idx_k = neighbors[:, :, idx_k]

    # Generate triple offset indices
    offset_idx_j = idx_j.repeat((B, A, 1))
    offset_idx_k = idx_k.repeat((B, A, 1))

    # Generate mask for triples
    mask_j = neighbor_mask[:, :, idx_j]
    mask_k = neighbor_mask[:, :, idx_k]
    mask_triples = mask_j * mask_k

    return nbh_idx_j, nhb_idx_k, offset_idx_j, offset_idx_k, mask_triples



#!/usr/bin/env python
import argparse
import logging
import os
import torch

import schnetpack as spk
import schnetpack.utils.script_utils.settings

logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))


class SpecialOption(argparse.Action):
    """
    Special argparse option. If option is not called, False is stored.
    If the option is called without argument, the value in const is provided as default.
    If the option is called with an argumeny, this argument is used as the value.
    """

    def __init__(
        self,
        option_strings,
        const,
        dest,
        type=type,
        default=False,
        required=False,
        help=None,
    ):
        super(SpecialOption, self).__init__(
            option_strings=option_strings,
            const=const,
            dest=dest,
            type=type,
            nargs="?",
            default=default,
            required=required,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)


def get_parser():
    """Setup parser for command line arguments"""
    main_parser = argparse.ArgumentParser()

    # general commands
    main_parser.add_argument("molecule_path", help="Initial geometry")
    main_parser.add_argument("model_path", help="Path of trained model")
    main_parser.add_argument("simulation_dir", help="Path to store MD data")
    main_parser.add_argument(
        "--device", help="Choose between 'cpu' and 'cuda'", default="cpu"
    )

    # environment settings
    main_parser.add_argument(
        "--environment_provider",
        type=str,
        default="simple",
        help="Environment provider (default: %(default)s)",
        choices=["simple", "ase", "torch"],
    )
    main_parser.add_argument(
        "--cutoff",
        type=float,
        default=5.0,
        help="Cutoff radius of local environment (default: %(default)s)",
    )

    # optimization
    main_parser.add_argument(
        "--optimize",
        const=1000,
        action=SpecialOption,
        type=int,
        help="Optimize geometry. A default of 1000 steps is used.",
    )

    # normal modes
    main_parser.add_argument(
        "--normal_modes", help="Compute normal modes", action="store_true"
    )

    # molecular dynamics options
    main_parser.add_argument(
        "--equilibrate",
        const=10000,
        action=SpecialOption,
        type=int,
        help="Equilibrate molecule for molecular dynamics. A default of 10000 steps is"
        "used.",
    )
    main_parser.add_argument(
        "--production",
        const=10000,
        action=SpecialOption,
        type=int,
        help="Production molecular dynamics (default: %(default)s)",
    )
    main_parser.add_argument(
        "--interval",
        type=int,
        default=100,
        help="Log trajectory every N steps (default: %(default)s)",
    )
    main_parser.add_argument(
        "--time_step",
        type=float,
        default=0.5,
        help="Timestep in fs (default: %(default)s)",
    )

    # temperature and bath options
    main_parser.add_argument(
        "--temp_init",
        type=float,
        default=300.0,
        help="Initial temperature used for molecular"
        " dynamics in K. (default: %(default)s)",
    )
    main_parser.add_argument(
        "--temp_bath",
        type=float,
        default=None,
        help="Temperature of bath. If None is given, NVE"
        " dynamics are run. (default: %(default)s)",
    )

    main_parser.add_argument(
        "--single_point",
        action="store_true",
        help="Perform a single point prediction on the "
        "current geometry (energies and forces).",
    )
    main_parser.add_argument(
        "--energy",
        type=str,
        help="Property name to the energy property in the dataset which has been "
        "used for training the model",
        default="energy",
    )
    main_parser.add_argument(
        "--forces",
        type=str,
        help="Property name to the forces property in the dataset which has been "
        "used for training the model",
        default="forces",
    )

    return main_parser


if __name__ == "__main__":

    parser = get_parser()
    args = parser.parse_args()
    argparse_dict = vars(args)
    jsonpath = os.path.join(args.simulation_dir, "args.json")

    device = torch.device(args.device)

    # Set up directory
    if not os.path.exists(args.simulation_dir):
        os.makedirs(args.simulation_dir)

    # Store command line args
    spk.utils.to_json(jsonpath, argparse_dict)

    # Load the model
    ml_model = spk.utils.load_model(args.model_path, map_location=device)
    logging.info("Loaded model.")

    logging.info(
        "The model you built has: {:d} parameters".format(
            spk.utils.count_params(ml_model)
        )
    )

    # Initialize the ML ase interface
    ml_calculator = spk.interfaces.AseInterface(
        args.molecule_path,
        ml_model,
        args.simulation_dir,
        args.device,
        args.energy,
        args.forces,
        environment_provider=spk.utils.get_environment_provider(args, device),
    )
    logging.info("Initialized ase driver")

    # Perform the requested simulations

    if args.single_point:
        logging.info("Single point prediction...")
        ml_calculator.calculate_single_point()

    if args.optimize:
        logging.info("Optimizing geometry...")
        ml_calculator.optimize(steps=args.optimize)

    if args.normal_modes:
        if not args.optimize:
            logging.warning(
                "Computing normal modes without optimizing the geometry makes me a sad "
                "schnetpack..."
            )
        logging.info("Computing normal modes...")
        ml_calculator.compute_normal_modes()

    if args.equilibrate:
        logging.info("Equilibrating the system...")
        if args.temp_bath is None:
            raise ValueError("Please supply bath temperature for equilibration")
        ml_calculator.init_md(
            "equilibration",
            time_step=args.time_step,
            interval=args.interval,
            temp_bath=args.temp_bath,
            temp_init=args.temp_init,
        )
        ml_calculator.run_md(args.equilibrate)

    if args.production:
        logging.info("Running production dynamics...")
        ml_calculator.init_md(
            "production",
            time_step=args.time_step,
            interval=args.interval,
            temp_bath=args.temp_bath,
            temp_init=args.temp_init,
        )
        ml_calculator.run_md(args.production)



"""
This module contains several parsers. This includes utilities for reading and converting molecular
dynamics input files to instructions for the :obj:`schnetpack.md.simulator.Simulator`. In addition, there is
a full package for parsing ORCA output files.
"""



import torch
import logging
import schnetpack as spk
from ase.data import atomic_numbers
import torch.nn as nn

__all__ = ["get_representation", "get_output_module", "get_model"]


def get_representation(args, train_loader=None):
    # build representation
    if args.model == "schnet":

        cutoff_network = spk.nn.cutoff.get_cutoff_by_string(args.cutoff_function)

        return spk.representation.SchNet(
            n_atom_basis=args.features,
            n_filters=args.features,
            n_interactions=args.interactions,
            cutoff=args.cutoff,
            n_gaussians=args.num_gaussians,
            cutoff_network=cutoff_network,
            normalize_filter=args.normalize_filter,
        )

    elif args.model == "wacsf":
        sfmode = ("weighted", "Behler")[args.behler]
        # Convert element strings to atomic charges
        elements = frozenset((atomic_numbers[i] for i in sorted(args.elements)))
        representation = spk.representation.BehlerSFBlock(
            args.radial,
            args.angular,
            zetas=set(args.zetas),
            cutoff_radius=args.cutoff,
            centered=args.centered,
            crossterms=args.crossterms,
            elements=elements,
            mode=sfmode,
        )
        logging.info(
            "Using {:d} {:s}-type SF".format(representation.n_symfuncs, sfmode)
        )
        # Standardize representation if requested
        if args.standardize:
            if train_loader is None:
                raise ValueError(
                    "Specification of a training_loader is required to standardize "
                    "wACSF"
                )
            else:
                logging.info("Computing and standardizing symmetry function statistics")
                return spk.representation.StandardizeSF(
                    representation, train_loader, cuda=args.cuda
                )

        else:
            return representation

    else:
        raise NotImplementedError("Unknown model class:", args.model)


def get_output_module_by_str(module_str):
    if module_str == "atomwise":
        return spk.atomistic.Atomwise
    elif module_str == "elemental_atomwise":
        return spk.atomistic.ElementalAtomwise
    elif module_str == "dipole_moment":
        return spk.atomistic.DipoleMoment
    elif module_str == "elemental_dipole_moment":
        return spk.atomistic.ElementalDipoleMoment
    elif module_str == "polarizability":
        return spk.atomistic.Polarizability
    elif module_str == "electronic_spatial_sxtent":
        return spk.atomistic.ElectronicSpatialExtent
    else:
        raise spk.utils.ScriptError(
            "{} is not a valid output " "module!".format(module_str)
        )


def get_output_module(args, representation, mean, stddev, atomref):
    derivative = spk.utils.get_derivative(args)
    negative_dr = spk.utils.get_negative_dr(args)
    contributions = spk.utils.get_contributions(args)
    stress = spk.utils.get_stress(args)
    if args.dataset == "md17" and not args.ignore_forces:
        derivative = spk.datasets.MD17.forces
    output_module_str = spk.utils.get_module_str(args)
    if output_module_str == "dipole_moment":
        return spk.atomistic.output_modules.DipoleMoment(
            args.features,
            predict_magnitude=True,
            mean=mean[args.property],
            stddev=stddev[args.property],
            property=args.property,
            contributions=contributions,
        )
    elif output_module_str == "electronic_spatial_extent":
        return spk.atomistic.output_modules.ElectronicSpatialExtent(
            args.features,
            mean=mean[args.property],
            stddev=stddev[args.property],
            property=args.property,
            contributions=contributions,
        )
    elif output_module_str == "atomwise":
        return spk.atomistic.output_modules.Atomwise(
            args.features,
            aggregation_mode=spk.utils.get_pooling_mode(args),
            mean=mean[args.property],
            stddev=stddev[args.property],
            atomref=atomref[args.property],
            property=args.property,
            derivative=derivative,
            negative_dr=negative_dr,
            contributions=contributions,
            stress=stress,
        )
    elif output_module_str == "polarizability":
        return spk.atomistic.output_modules.Polarizability(
            args.features,
            aggregation_mode=spk.utils.get_pooling_mode(args),
            property=args.property,
        )
    elif output_module_str == "isotropic_polarizability":
        return spk.atomistic.output_modules.Polarizability(
            args.features,
            aggregation_mode=spk.utils.get_pooling_mode(args),
            property=args.property,
            isotropic=True,
        )
    # wacsf modules
    elif output_module_str == "elemental_dipole_moment":
        elements = frozenset((atomic_numbers[i] for i in sorted(args.elements)))
        return spk.atomistic.output_modules.ElementalDipoleMoment(
            representation.n_symfuncs,
            n_hidden=args.n_nodes,
            n_layers=args.n_layers,
            predict_magnitude=True,
            elements=elements,
            property=args.property,
        )
    elif output_module_str == "elemental_atomwise":
        elements = frozenset((atomic_numbers[i] for i in sorted(args.elements)))
        return spk.atomistic.output_modules.ElementalAtomwise(
            representation.n_symfuncs,
            n_hidden=args.n_nodes,
            n_layers=args.n_layers,
            aggregation_mode=spk.utils.get_pooling_mode(args),
            mean=mean[args.property],
            stddev=stddev[args.property],
            atomref=atomref[args.property],
            elements=elements,
            property=args.property,
            derivative=derivative,
            negative_dr=negative_dr,
        )
    else:
        raise NotImplementedError


def get_model(args, train_loader, mean, stddev, atomref, logging=None):
    """
    Build a model from selected parameters or load trained model for evaluation.

    Args:
        args (argsparse.Namespace): Script arguments
        train_loader (spk.AtomsLoader): loader for training data
        mean (torch.Tensor): mean of training data
        stddev (torch.Tensor): stddev of training data
        atomref (dict): atomic references
        logging: logger

    Returns:
        spk.AtomisticModel: model for training or evaluation
    """
    if args.mode == "train":
        if logging:
            logging.info("building model...")
        representation = get_representation(args, train_loader)
        output_module = get_output_module(
            args,
            representation=representation,
            mean=mean,
            stddev=stddev,
            atomref=atomref,
        )
        model = spk.AtomisticModel(representation, [output_module])

        if args.parallel:
            model = nn.DataParallel(model)
        if logging:
            logging.info(
                "The model you built has: %d parameters" % spk.utils.count_params(model)
            )
        return model
    else:
        raise spk.utils.ScriptError("Invalid mode selected: {}".format(args.mode))



import logging
import os
import shutil
import tarfile
import tempfile
import numpy as np
from urllib import request as request
from urllib.error import HTTPError, URLError

import schnetpack as spk
from schnetpack.datasets import DownloadableAtomsData


class ISO17(DownloadableAtomsData):
    """
    ISO17 benchmark data set for molecular dynamics of C7O2H10 isomers
    containing molecular forces.

    Args:
        datapath (str): Path to database directory
        fold (str): Fold of data to load. Allowed are:
                        reference - 80% of steps of 80% of MD trajectories
                        reference_eq - equilibrium conformations of those
                                       molecules
                        test_within - remaining 20% unseen steps of reference trajectories
                        test_other - remaining 20% unseen MD trajectories
                        test_eq - equilibrium conformations of test trajectories
        subset (list, optional): Deprecated! Do not use! Subsets are created with
            AtomsDataSubset class.
        load_only (list, optional): reduced set of properties to be loaded
        download (bool): set to true if dataset should be downloaded. (default: True)
        collect_triples (false): set to true to compute triples for angular functions
            (default: true)
        environment_provider (spk.environment.BaseEnvironmentProvider): define how
            neighborhood is calculated
            (default=spk.environment.SimpleEnvironmentProvider).

    See: http://quantum-machine.org/datasets/
    """

    existing_folds = [
        "reference",
        "reference_eq",
        "test_within",
        "test_other",
        "test_eq",
    ]

    E = "total_energy"
    F = "atomic_forces"

    def __init__(
        self,
        datapath,
        fold,
        download=True,
        load_only=None,
        subset=None,
        collect_triples=False,
        environment_provider=spk.environment.SimpleEnvironmentProvider(),
    ):

        if fold not in self.existing_folds:
            raise ValueError("Fold {:s} does not exist".format(fold))

        available_properties = [ISO17.E, ISO17.F]
        units = [1.0, 1.0]

        self.path = datapath
        self.fold = fold
        dbpath = os.path.join(datapath, "iso17", fold + ".db")

        super().__init__(
            dbpath=dbpath,
            subset=subset,
            load_only=load_only,
            collect_triples=collect_triples,
            download=download,
            available_properties=available_properties,
            units=units,
            environment_provider=environment_provider,
        )

    def _download(self):
        logging.info("Downloading ISO17 database...")
        tmpdir = tempfile.mkdtemp("iso17")
        tarpath = os.path.join(tmpdir, "iso17.tar.gz")
        url = "http://www.quantum-machine.org/datasets/iso17.tar.gz"

        try:
            request.urlretrieve(url, tarpath)
        except HTTPError as e:
            logging.error("HTTP Error:", e.code, url)
            return False
        except URLError as e:
            logging.error("URL Error:", e.reason, url)
            return False

        tar = tarfile.open(tarpath)
        tar.extractall(self.path)
        tar.close()

        shutil.rmtree(tmpdir)

        return True



"""
Class for extracting information from the HDF5 files generated during simulation by the
:obj:`schnetpack.md.simulation_hooks.logging_hooks.FileLogger`.
In addition to loading structures, velocities, etc., various postprocessing functions are available.
"""

import json
import logging
import h5py
import numpy as np
from ase import data
import warnings

from schnetpack import Properties
from schnetpack.md.utils import MDUnits


class HDF5LoaderError(Exception):
    """
    Exception for HDF5 loader class.
    """

    pass


class HDF5Loader:
    """
    Class for loading HDF5 datasets written by the FileLogger. By default, this requires at least a MoleculeSteam to be
    present. PropertyData is also read by default, but can be disabled.

    Args:
        hdf5_database (str): Path to the database file.
        skip_initial (int): Skip the initial N configurations in the trajectory, e.g. to account for equilibration
                            (default=0).
        load_properties (bool): Extract and reconstruct the property data stored by a PropertyStream (e.g. forces,
                                energies, etc.), enabled by default.
    """

    def __init__(self, hdf5_database, skip_initial=0, load_properties=True):
        self.database = h5py.File(hdf5_database, "r", swmr=True, libver="latest")
        self.skip_initial = skip_initial
        self.data_groups = list(self.database.keys())

        self.properties = {}

        # Load basic structure properties and MD info
        if "molecules" not in self.data_groups:
            raise HDF5LoaderError(
                "Molecule data not found in {:s}".format(hdf5_database)
            )
        else:
            self._load_molecule_data()

        # If requested, load other properties predicted by the model stored via PropertyStream
        if load_properties:
            if "properties" not in self.data_groups:
                raise HDF5LoaderError(
                    "Molecule properties not found in {:s}".format(hdf5_database)
                )
            else:
                self._load_properties()

        # Do formatting for info
        loaded_properties = list(self.properties.keys())
        if len(loaded_properties) == 1:
            loaded_properties = str(loaded_properties[0])
        else:
            loaded_properties = (
                ", ".join(loaded_properties[:-1]) + " and " + loaded_properties[-1]
            )

        logging.info(
            "Loaded properties {:s} from {:s}".format(loaded_properties, hdf5_database)
        )

    def _load_molecule_data(self):
        """
        Load data stored by a MoleculeStream. This contains basic information on the system and is required to be
        present for the Loader.
        """
        # This is for molecule streams
        structures = self.database["molecules"]

        # General database info
        self.n_replicas = structures.attrs["n_replicas"]
        self.n_molecules = structures.attrs["n_molecules"]
        self.n_atoms = structures.attrs["n_atoms"]
        self.total_entries = structures.attrs["entries"]
        self.time_step = structures.attrs["time_step"]
        self.entries = self.total_entries - self.skip_initial

        # Write to main property dictionary
        if structures.ndim == 4:
            self._load_structures(structures)
        else:
            # Check for deprecated format
            warnings.warn(
                "Old format of hdf5 detected, periodic systems not supported.",
                DeprecationWarning,
            )
            self._load_structures_deprectated(structures)

    def _load_structures_deprectated(self, structures):
        self.properties[Properties.Z] = structures.attrs["atom_types"][0]

        # Get positions
        self.properties[Properties.R] = structures[
            self.skip_initial : self.total_entries, ..., :3
        ]
        # Get velocities
        self.properties["velocities"] = structures[
            self.skip_initial : self.total_entries, ..., 3:
        ]

    def _load_structures(self, structures):
        self.properties[Properties.Z] = structures.attrs["atom_types"][0]

        # Get length of position and velocity blocks
        max_n_atoms = structures.attrs["max_n_atoms"]
        block_length = 3 * max_n_atoms
        self.pbc = structures.attrs["pbc"]

        # Get positions
        self.properties[Properties.R] = structures[
            self.skip_initial : self.total_entries, ..., :block_length
        ].reshape((self.entries, self.n_replicas, self.n_molecules, max_n_atoms, 3))
        # Get velocities
        self.properties["velocities"] = structures[
            self.skip_initial : self.total_entries, ..., block_length : 2 * block_length
        ].reshape((self.entries, self.n_replicas, self.n_molecules, max_n_atoms, 3))

        # Get cells if present
        if structures.shape[-1] > 2 * block_length:
            self.properties[Properties.cell] = structures[
                self.skip_initial : self.total_entries, ..., 2 * block_length :
            ].reshape((self.entries, self.n_replicas, self.n_molecules, 3, 3))
        else:
            self.properties[Properties.cell] = None

    def _load_properties(self):
        """
        Load properties and their shape from the corresponding group in the hdf5 database.
        Properties are then reshaped to original form and stores in the self.properties dictionary.
        """
        # And for property stream
        properties = self.database["properties"]
        property_shape = json.loads(properties.attrs["shapes"])
        property_positions = json.loads(properties.attrs["positions"])

        # Reformat properties
        for prop in property_positions:
            prop_pos = slice(*property_positions[prop])
            self.properties[prop] = properties[
                self.skip_initial : self.total_entries, :, :, prop_pos
            ].reshape(
                (
                    self.total_entries - self.skip_initial,
                    self.n_replicas,
                    self.n_molecules,
                    *property_shape[prop],
                )
            )

    def get_property(self, property_name, mol_idx=0, replica_idx=None, atomistic=False):
        """
        Extract property from dataset.

        Args:
            property_name (str): Name of the property as contained in the self.properties dictionary.
            mol_idx (int): Index of the molecule to extract, by default uses the first molecule (mol_idx=0)
            replica_idx (int): Replica of the molecule to extract (e.g. for ring polymer molecular dynamics). If
                               replica_idx is set to None (default), the centroid is returned if multiple replicas are
                               present.
            atomistic (bool): Whether the property is atomistic (e.g. forces) or defined for the whole molecule
                              (e.g. energies, dipole moments). If set to True, the array is masked according to the
                               number of atoms for the requested molecule to counteract potential zero-padding.
                               (default=False)

        Returns:
            np.array: N_steps x property dimensions array containing the requested property collected during the simulation.
        """

        # Special case for atom types
        if property_name == Properties.Z:
            return self.properties[Properties.Z][mol_idx]

        # Check whether property is present
        if property_name not in self.properties:
            raise HDF5LoaderError(f"Property {property_name} not found in database.")

        # Mask by number of atoms if property is declared atomistic
        if atomistic:
            n_atoms = self.n_atoms[mol_idx]
            target_property = self.properties[property_name][
                :, :, mol_idx, :n_atoms, ...
            ]
        else:
            if self.properties[property_name] is not None:
                target_property = self.properties[property_name][:, :, mol_idx, ...]
            else:
                target_property = None

        # Compute the centroid unless requested otherwise
        if replica_idx is None:
            if target_property is not None:
                target_property = np.mean(target_property, axis=1)
        else:
            if target_property is not None:
                target_property = target_property[:, replica_idx, ...]

        return target_property

    def get_velocities(self, mol_idx=0, replica_idx=None):
        """
        Auxiliary routine for getting the velocities of specific molecules and replicas.

        Args:
            mol_idx (int): Index of the molecule to extract, by default uses the first molecule (mol_idx=0)
            replica_idx (int): Replica of the molecule to extract (e.g. for ring polymer molecular dynamics). If
                               replica_idx is set to None (default), the centroid is returned if multiple replicas are
                               present.

        Returns:
            np.array: N_steps x N_atoms x 3 array containing the atom velocities of the simulation in atomic units.
        """
        return self.get_property(
            "velocities", mol_idx=mol_idx, replica_idx=replica_idx, atomistic=True
        )

    def get_positions(self, mol_idx=0, replica_idx=None):
        """
        Auxiliary routine for getting the positions of specific molecules and replicas.

        Args:
            mol_idx (int): Index of the molecule to extract, by default uses the first molecule (mol_idx=0)
            replica_idx (int): Replica of the molecule to extract (e.g. for ring polymer molecular dynamics). If
                               replica_idx is set to None (default), the centroid is returned if multiple replicas are
                               present.

        Returns:
            np.array: N_steps x N_atoms x 3 array containing the atom positions of the simulation in atomic units.
        """
        return self.get_property(
            Properties.R, mol_idx=mol_idx, replica_idx=replica_idx, atomistic=True
        )

    def get_kinetic_energy(self, mol_idx=0, replica_idx=None):
        """
        Auxiliary routine for computing the kinetic energy of every configuration based on it's velocities.

        Args:
            mol_idx (int): Index of the molecule to extract, by default uses the first molecule (mol_idx=0)
            replica_idx (int): Replica of the molecule to extract (e.g. for ring polymer molecular dynamics). If
                               replica_idx is set to None (default), the centroid is returned if multiple replicas are
                               present.

        Returns:
            np.array: N_steps array containing the kinetic eenrgy of every configuration in atomic units.
        """
        # Get the velocities
        velocities = self.get_velocities(mol_idx=mol_idx, replica_idx=replica_idx)

        # Get the masses and convert to correct units
        masses = (
            data.atomic_masses[self.properties[Properties.Z][mol_idx]]
            * MDUnits.da2internal
        )

        # Compute the kinetic energy as 1/2*m*v^2
        kinetic_energy = 0.5 * np.sum(
            masses[None, :, None] * velocities ** 2, axis=(1, 2)
        )
        return kinetic_energy

    def get_temperature(self, mol_idx=0, replica_idx=None):
        """
        Auxiliary routine for computing the instantaneous temperature of every configuration.

        Args:
            mol_idx (int): Index of the molecule to extract, by default uses the first molecule (mol_idx=0)
            replica_idx (int): Replica of the molecule to extract (e.g. for ring polymer molecular dynamics). If
                               replica_idx is set to None (default), the centroid is returned if multiple replicas are
                               present.

        Returns:
            np.array: N_steps array containing the temperature of every configuration in Kelvin.
        """
        # Get the velocities
        # Get the kinetic energy
        kinetic_energy = self.get_kinetic_energy(
            mol_idx=mol_idx, replica_idx=replica_idx
        )

        # Compute the temperature
        temperature = 2.0 / (3.0 * MDUnits.kB * self.n_atoms[mol_idx]) * kinetic_energy

        return temperature



import os
import numpy as np
from schnetpack.data import AtomsDataError, AtomsDataSubset


def train_test_split(
    data,
    num_train=None,
    num_val=None,
    split_file=None,
    stratify_partitions=False,
    num_per_partition=False,
):
    """
    Splits the dataset into train/validation/test splits, writes split to
    an npz file and returns subsets. Either the sizes of training and
    validation split or an existing split file with split indices have to
    be supplied. The remaining data will be used in the test dataset.

    Args:
        data (spk.data.AtomsData): full atomistic dataset
        num_train (int): number of training examples
        num_val (int): number of validation examples
        split_file (str): Path to split file. If file exists, splits will
                          be loaded. Otherwise, a new file will be created
                          where the generated split is stored.

    Returns:
        spk.data.AtomsDataSubset: subset with training data
        spk.data.AtomsDataSubset: subset with validation data
        spk.data.AtomsDataSubset: subset with test data

    """
    if split_file is not None and os.path.exists(split_file):
        S = np.load(split_file)
        train_idx = S["train_idx"].tolist()
        val_idx = S["val_idx"].tolist()
        test_idx = S["test_idx"].tolist()
    else:
        if num_train is None or num_val is None:
            raise ValueError(
                "You have to supply either split sizes (num_train /"
                + " num_val) or an npz file with splits."
            )

        assert num_train + num_val <= len(
            data
        ), "Dataset is smaller than num_train + num_val!"

        num_train = num_train if num_train > 1 else num_train * len(data)
        num_val = num_val if num_val > 1 else num_val * len(data)
        num_train = int(num_train)
        num_val = int(num_val)

        if stratify_partitions:
            partitions = data.get_metadata("partitions")
            n_partitions = len(partitions)
            if num_per_partition:
                num_train_part = num_train
                num_val_part = num_val
            else:
                num_train_part = num_train // n_partitions
                num_val_part = num_val // n_partitions

            train_idx = []
            val_idx = []
            test_idx = []
            for start, stop in partitions.values():
                idx = np.random.permutation(np.arange(start, stop))
                train_idx += idx[:num_train_part].tolist()
                val_idx += idx[num_train_part : num_train_part + num_val_part].tolist()
                test_idx += idx[num_train_part + num_val_part :].tolist()

        else:
            idx = np.random.permutation(len(data))
            train_idx = idx[:num_train].tolist()
            val_idx = idx[num_train : num_train + num_val].tolist()
            test_idx = idx[num_train + num_val :].tolist()

        if split_file is not None:
            np.savez(
                split_file, train_idx=train_idx, val_idx=val_idx, test_idx=test_idx
            )

    train = create_subset(data, train_idx)
    val = create_subset(data, val_idx)
    test = create_subset(data, test_idx)

    return train, val, test


def create_subset(dataset, indices):
    r"""
    Create a subset of atomistic datasets.

    Args:
        dataset (torch.utils.data.Dataset): dataset
        indices (sequence): indices of the subset; no np.ndarrays, because the ase database can not handle np.int values

    Returns:
        spk.data.AtomsDataSubset: subset of input dataset

    """
    max_id = 0 if len(indices) == 0 else max(indices)
    if len(dataset) <= max_id:
        raise AtomsDataError(
            "The subset indices do not match the total length of the dataset!"
        )
    return AtomsDataSubset(dataset, indices)



"""
This module contains all routines for influencing the sampling during a molecular dynamics run.
In general, these routines are derived from :obj:`schnetpack.md.simulation_hooks.SimulationHook` and act
on the middle part of each simulator step.
Currently, accelerated molecular dynamics and metadynamics are implemented.
"""
import numpy as np
import copy
import logging
import os

import torch
from schnetpack import AtomsData
from schnetpack.md import MaxwellBoltzmannInit

from schnetpack.md.simulation_hooks import SimulationHook
from schnetpack.md.utils import MDUnits

__all__ = [
    "BiasPotential",
    "AcceleratedMD",
    "MetaDyn",
    "CollectiveVariable",
    "BondColvar",
    "AdaptiveSampling",
]


class BiasPotential(SimulationHook):
    """
    Placeholder class for bias potentials used for adaptive/accelerated sampling.
    """

    def __init__(self):
        raise NotImplementedError

    def on_step_end(self, simulator):
        """
        Needs to be implemented. This applies a bias potential after computing the forces
        and properties.

        Args:
            simulator (schnetpack.md.Simulator): Main simulator used for driving the dynamics
        """
        raise NotImplementedError


class AcceleratedMD(SimulationHook):
    """
    Hook for performing accelerated molecular dynamics [#accmd1]_ . This method distorts the potential
    energy surface in order to make deep valleys smoother. This smoothing is applied to everything
    below an energy threshold and its strength is regulated via a acceleration factor.
    The energy conversion can be used to specify the energy threshold in arbitrary units of energy, which are then
    converted to the internal units.

    Args:
        energy_threshold (float): Energy threshold in units of energy used by the calculator.
        acceleration_factor (float): Acceleration factor steering the smoothness of the bias potential
        energy_handle (str, optional): Identifier for the potential energies.
        energy_conversion (str, optional): Conversion factor for the calculator energies.

    References
    ----------
    .. [#accmd1] Hamelberg, Mongan, McCammon:
                 Accelerated molecular dynamics: a promising and efficient simulation method for biomolecules.
                 J. Chem. Phys., 120 (24), 11919--11929. 2004.
    """

    def __init__(
        self,
        energy_threshold,
        acceleration_factor,
        energy_handle="energy",
        energy_conversion=1.0,
    ):
        self.energy_threshold = energy_threshold
        self.acceleration_factor = acceleration_factor
        self.energy_handle = energy_handle
        # Convert from calculator -> internal
        self.energy_conversion = MDUnits.unit2internal(energy_conversion)

    def on_step_middle(self, simulator):
        """
        Compute the bias potential and derivatives and use them to update the
        current state of :obj:`schnetpack.md.System` in the simulator.
        While the forces are updated, the bias potential itself is stored in the
        properties dictionary of the system.

        Args:
            simulator (schnetpack.md.Simulator): Main simulator used for driving the dynamics
        """
        energies, forces = self._get_energies_forces(simulator)

        # Compute bias potential and derivatives
        bias_potential = self._function(energies)
        bias_forces = self._derivative(energies) * forces

        # Store bias potential and update forces
        simulator.system.properties["bias_potential"] = bias_potential
        simulator.system.forces = forces + bias_forces.detach()

    def _get_energies_forces(self, simulator):
        """
        Extract energies and forces from the present system and
        convert them to the units used by the bias potential.

        Args:
            simulator (schnetpack.md.Simulator): Main simulator used for driving the dynamics

        Returns:
            (torch.Tensor,torch.Tensor):
                energies:
                    torch.Tensor containing the system energies (in calculator units)

                forces:
                    torch.Tensor containing the system forces (in atomic units)

        """
        energies = (
            simulator.system.properties[self.energy_handle].float()
            * self.energy_conversion
        )
        forces = simulator.system.forces
        return energies, forces

    def _function(self, energies):
        """
        Primary bias potential function used in accelerated molecular dynamics.

        Args:
            energies (torch.Tensor): Potential energies to which the bias potential
                                     should be applied to.

        Returns:
            torch.Tensor: Computed bias function.
        """
        function = torch.pow(self.energy_threshold - energies, 2.0) / (
            self.acceleration_factor + self.energy_threshold - energies
        )
        function[energies >= self.energy_threshold] = 0
        return function

    def _derivative(self, energies):
        """
        Compute the derivative of the bias function for updating the forces.
        Effectively, this becomes a scaling factor of the original forces.

        Args:
            energies (torch.Tensor): Potential energies to which the bias potential
                                     should be applied to.

        Returns:
            torch.Tensor: Derivative of the bias function.
        """
        derivative = (
            self.acceleration_factor ** 2
            / torch.pow(
                self.acceleration_factor + self.energy_threshold - energies, 2.0
            )
            - 1.0
        )
        derivative[energies >= self.energy_threshold] = 0
        return derivative


class CollectiveVariable:
    """
    Basic collective variable to be used in combination with the :obj:`MetaDyn` hook.
    The ``_colvar_function`` needs to be implemented based on the desired collective
    variable.

    Args:
        width (float): Parameter regulating the standard deviation of the Gaussians.
    """

    def __init__(self, width):
        # Initialize the width of the Gaussian
        self.width = 0.5 * width ** 2

    def get_colvar(self, structure):
        """
        Compute the collecyive variable.

        Args:
            structure (torch.Tensor): Atoms positions taken from the system in the :obj:`schnetpack.md.Simulator`.

        Returns:
            torch.Tensor: Collective variable computed for the structure.
        """
        return self._colvar_function(structure)

    def _colvar_function(self, structure):
        """
        Placeholder for defining the particular collective variable function to be used.

        Args:
            structure (torch.Tensor): Atoms positions taken from the system in the :obj:`schnetpack.md.Simulator`.
        """
        raise NotImplementedError


class BondColvar(CollectiveVariable):
    """
    Collective variable acting on bonds between atoms.
    ``idx_a`` indicates the index of the first atom in the structure tensor, ``idx_b`` the second.
    Counting starts at zero.

    Args:
        idx_a (int): Index of the first atom in the positions tensor provided by the simulator system.
        idx_b (int): Index of the second atom.
        width (float): Width of the Gaussians applied to this collective variable. For bonds, units of Bohr are used.
    """

    def __init__(self, idx_a, idx_b, width):
        super(BondColvar, self).__init__(width)
        self.idx_a = idx_a
        self.idx_b = idx_b

    def _colvar_function(self, structure):
        """
        Compute the distance between both atoms.

        Args:
            structure (torch.Tensor): Atoms positions taken from the system in the :obj:`schnetpack.md.Simulator`.

        Returns:
            torch.Tensor: Bind collective variable.
        """
        vector_ab = structure[:, :, self.idx_b, :] - structure[:, :, self.idx_a, :]
        return torch.norm(vector_ab, 2, dim=2)


class MetaDyn(SimulationHook):
    """
    Perform a metadynamics simulation, where Gaussian potentials are deposited along collective
    variables in order to steer the sampling of a molecular dynamics trajectory. [#metadyn1]_

    Args:
        collective_variables (list): List of collective variables to be sampled
                                     (:obj:`schnetpack.md.simulation_hooks.CollectiveVariable`).
        frequency (int, optional): Frequency with which Gaussians are deposited (every n simulation steps).
        weight (float, optional): Weight of each Gaussian in units of energy (Hartee).
        store_potential (bool, optional): Whether centers and widths of the placed Gaussians should be stored.

    References
    ----------
    .. [#metadyn1] Laio, Parrinello:
                   Escaping free-energy minima.
                   Proc. Natl. Acad. Sci., 99 (20), 12562--12566, 2002.

    """

    def __init__(
        self,
        collective_variables,
        frequency=200,
        weight=1.0 / 627.509,
        store_potential=True,
    ):
        self.collective_variables = collective_variables
        self.store_potential = store_potential

        self.gaussian_centers = None
        self.gaussian_mask = None
        self.collective_variable_widths = None

        self.frequency = frequency
        self.weigth = weight
        self.n_gaussians = 0

    def on_simulation_start(self, simulator):
        """
        Initialize the tensor holding the Gaussian centers and widths. These
        will be populated during the simulation.

        Args:
            simulator (schnetpack.md.Simulator): Main simulator used for driving the dynamics
        """
        n_gaussian_centers = int(simulator.n_steps / self.frequency) + 1
        self.gaussian_centers = torch.zeros(
            n_gaussian_centers,
            len(self.collective_variables),
            device=simulator.system.device,
        )
        self.collective_variable_widths = torch.FloatTensor(
            [cv.width for cv in self.collective_variables],
            device=simulator.system.device,
        )
        self.gaussian_mask = torch.zeros(
            n_gaussian_centers, device=simulator.system.device
        )
        self.gaussian_mask[0] = 1

    def on_step_middle(self, simulator):
        """
        Based on the current structure, compute the collective variables and the
        associated Gaussian potentials. If multiple collective variables are given,
        a product potential is formed. torch.autograd is used to compute the forces
        resulting from the potential, which are then in turn used to update the system
        forces. A new Gaussian is added after a certain number of steps.

        Args:
            simulator (schnetpack.md.Simulator): Main simulator used for driving the dynamics
        """
        # Get and detach the structure from the simulator
        structure = simulator.system.positions.detach()
        # Enable gradients for bias forces
        structure.requires_grad = True

        # Compute the collective variables
        colvars = torch.stack(
            [colvar.get_colvar(structure) for colvar in self.collective_variables],
            dim=2,
        )

        # Compute the Gaussians for the potential
        gaussians = torch.exp(
            -((colvars[:, :, None, :] - self.gaussian_centers[None, None, :, :]) ** 2)
            / self.collective_variable_widths[None, None, None, :]
        )
        # Compute the bias potential and apply mask for centers not yet stored
        bias_potential = torch.prod(gaussians, dim=3) * self.gaussian_mask

        # Finalize potential and compute forces
        bias_potential = torch.sum(self.weigth * bias_potential, 2)
        bias_forces = -torch.autograd.grad(
            bias_potential, structure, torch.ones_like(bias_potential)
        )[0]

        # Store bias potential, collective variables and update system forces
        simulator.system.properties["bias_potential"] = bias_potential.detach()
        simulator.system.properties["collective_variables"] = colvars.detach()
        simulator.system.forces = simulator.system.forces + bias_forces.detach()

        if self.store_potential:
            # TODO: Much better to move this to a state dict?
            # Store information on the general shape of the bias potential
            simulator.system.properties[
                "gaussian_centers"
            ] = self.gaussian_centers.detach()
            simulator.system.properties[
                "gaussian_widths"
            ] = self.collective_variable_widths.detach()
            simulator.system.properties["gaussian_mask"] = self.gaussian_mask.detach()

        # Add a new Gaussian to the potential every n_steps
        if simulator.step % self.frequency == 0:
            self.gaussian_centers[self.n_gaussians] = colvars.detach()
            # Update the mask
            self.gaussian_mask[self.n_gaussians + 1] = 1
            self.n_gaussians += 1


class AdaptiveSampling(SimulationHook):
    def __init__(
        self,
        thresholds,
        n_samples,
        dataset,
        reset=True,
        temperature=300,
        initializer=MaxwellBoltzmannInit,
    ):

        self.thresholds = thresholds
        self.n_samples = n_samples
        self.samples = []
        self.samples_thresholds = []

        # Reinitialization
        self.reset = reset
        self.initializer = initializer(temperature)

        # Dataset for storage
        self.dataset = dataset
        if os.path.exists(self.dataset):
            logging.info(
                "Database {:s} already exists. Data will be appended.".format(
                    self.dataset
                )
            )

        # Initial system variables for reset
        self.init_positions = None
        self.init_forces = None
        self.init_cells = None

    def on_simulation_start(self, simulator):
        # Store initial configs for system reset
        self.init_positions = copy.deepcopy(simulator.system.positions)
        self.init_forces = copy.deepcopy(simulator.system.forces)
        self.init_cells = copy.deepcopy(simulator.system.cells)

    def on_step_end(self, simulator):
        # Check if a sample is required
        sample_system, sample_molecule, threshold_exceeded = self._check_uncertainty(
            simulator.system
        )

        if sample_system:
            # Collect samples based on uncertainty thresholds
            self._collect_samples(simulator.system, sample_molecule, threshold_exceeded)

            # Reinitialize velocities if requested
            if self.reset:
                logging.info("Resetting system...")
                self._reset_system(simulator.system, mask=sample_molecule)

        # If a sufficient number of samples is collected, stop sampling
        if len(self.samples) >= self.n_samples:
            self._write_database()
            exit()

    def on_simulation_end(self, simulator):
        # Wrap everything up if simulation finished without collecting all/any samples.
        self._write_database()

    def _check_uncertainty(self, system):
        threshold_exceeded = {}

        sample_system = False
        sample_molecule = torch.zeros(
            system.n_replicas, system.n_molecules, device=system.device
        ).bool()

        # Check if a sample is needed
        for prop in self.thresholds:
            # Get variance from simulator
            prop_var = system.properties["{:s}_var".format(prop)]

            # Reshaping depending if property is atomic or not
            shapes = prop_var.shape
            if shapes[2] == system.max_n_atoms:
                prop_var.view(*shapes[:2], -1)
                uncertainty = torch.sqrt(torch.sum(prop_var, dim=-1))
            else:
                prop_var.view(*shapes[:1], -1)
                uncertainty = torch.sqrt(torch.sum(prop_var, dim=-1, keepdim=True))

            # Check if uncertainty threshold is exceeded
            threshold_exceeded[prop] = self.thresholds[prop] < uncertainty

            # Checks if a) a sample is needed, b) for which replica/molecule a sample is needed, c) for which atom
            if torch.any(threshold_exceeded[prop]):
                # Check for which molecule/replica samples are required
                sample_molecule = sample_molecule | torch.any(
                    threshold_exceeded[prop], dim=-1
                )
                # Overall sample required
                sample_system = True

        return sample_system, sample_molecule, threshold_exceeded

    def _collect_samples(self, system, sample_molecule, threshold_exceeded):
        # Get structures in the form of ASE atoms (R x M is flattened)
        atoms = system.get_ase_atoms(internal_units=False)

        # Collect all replicas and molecules which need sampling
        idx_c = 0
        for idx_r in range(system.n_replicas):
            for idx_m in range(system.n_molecules):

                # Get the atoms and store the thresholds
                if sample_molecule[idx_r, idx_m] == 1:
                    self.samples.append(atoms[idx_c])
                    sample_thresholds = {}
                    for prop in threshold_exceeded:
                        sample_thresholds[prop] = (
                            threshold_exceeded[prop][idx_r, idx_m]
                            .detach()
                            .cpu()
                            .numpy()
                            .astype(np.float32)
                        )
                    self.samples_thresholds.append(sample_thresholds)
                idx_c += 1

    def _reset_system(self, system, mask=None):
        if mask is None:
            system.positions = copy.deepcopy(self.init_positions)
            system.forces = copy.deepcopy(self.init_forces)
            system.cells = copy.deepcopy(self.init_cells)
            self.initializer.initialize_system(system)
        else:
            system.positions[mask == 1, ...] = copy.deepcopy(
                self.init_positions[mask == 1, ...]
            )
            system.forces[mask == 1, ...] = copy.deepcopy(
                self.init_forces[mask == 1, ...]
            )
            if self.init_cells is not None:
                system.cells[mask == 1, ...] = copy.deepcopy(
                    self.init_cells[mask == 1, ...]
                )
            self.initializer.initialize_system(system, mask=mask)

    def _write_database(self):
        if len(self.samples) > 0:
            dataset = AtomsData(
                self.dataset, available_properties=self.samples_thresholds[0]
            )
            dataset.add_systems(self.samples, self.samples_thresholds)
            logging.info(
                "{:d} samples written to {:s}.".format(len(self.samples), self.dataset)
            )
        else:
            logging.info("No samples collected.")



import logging
import torch


logger = logging.getLogger(__name__)


class StatisticsAccumulator:
    def __init__(self, batch=False, atomistic=False):
        """
        Use the incremental Welford algorithm described in [1]_ to accumulate
        the mean and standard deviation over a set of samples.

        Args:
            batch: If set to true, assumes sample is batch and uses leading
                   dimension as batch size
            atomistic: If set to true, average over atom dimension

        References:
        -----------
        .. [1] https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance

        """
        # Initialize state variables
        self.count = 0  # Sample count
        self.mean = 0  # Incremental average
        self.M2 = 0  # Sum of squares of differences
        self.batch = batch
        self.atomistic = atomistic

    def add_sample(self, sample_value):
        """
        Add a sample to the accumulator and update running estimators.
        Differentiates between different types of samples.

        Args:
            sample_value (torch.Tensor): data sample
        """

        # Check different cases
        if not self.batch and not self.atomistic:
            self._add_sample(sample_value)
        elif not self.batch and self.atomistic:
            n_atoms = sample_value.size(0)
            for i in range(n_atoms):
                self._add_sample(sample_value[i, :])
        elif self.batch and not self.atomistic:
            n_batch = sample_value.size(0)
            for i in range(n_batch):
                self._add_sample(sample_value[i, :])
        else:
            n_batch = sample_value.shape[0]
            n_atoms = sample_value.shape[1]
            for i in range(n_batch):
                for j in range(n_atoms):
                    self._add_sample(sample_value[i, j, :])

    def _add_sample(self, sample_value):
        # Update count
        self.count += 1
        delta_old = sample_value - self.mean
        # Difference to old mean
        self.mean += delta_old / self.count
        # Update mean estimate
        delta_new = sample_value - self.mean
        # Update sum of differences
        self.M2 += delta_old * delta_new

    def get_statistics(self):
        """
        Compute statistics of all data collected by the accumulator.

        Returns:
            torch.Tensor: Mean of data
            torch.Tensor: Standard deviation of data
        """
        # Compute standard deviation from M2
        mean = self.mean
        stddev = torch.sqrt(self.M2 / self.count)

        return mean, stddev

    def get_mean(self):
        return self.mean

    def get_stddev(self):
        return torch.sqrt(self.M2 / self.count)
